import os

from ase.io import read

from TB2J.myTB import MyTB
from TB2J.utils import auto_assign_basis_name
from TB2J.wannier import parse_atoms

from .manager import Manager


class WannierManager(Manager):
    def __init__(
        self,
        path,
        prefix_up,
        prefix_dn,
        prefix_SOC,
        colinear=True,
        groupby="spin",
        posfile=None,
        basis_fname=None,
        qspace=False,
        wannier_type="wannier90",
        **kwargs,
    ):
        # atoms
        atoms = self.prepare_atoms(path, posfile, prefix_up, prefix_SOC, colinear)
        output_path = kwargs.get("output_path", "TB2J_results")
        # models and basis
        if colinear:
            tbmodels, basis = self.prepare_model_colinear(
                path, prefix_up, prefix_dn, atoms, output_path=output_path
            )
        else:
            tbmodels, basis = self.prepare_model_ncl(
                path, prefix_SOC, atoms, groupby, output_path=output_path
            )

        description = self.description(path, prefix_up, prefix_dn, prefix_SOC, colinear)
        kwargs["description"] = description

        super().__init__(atoms, tbmodels, basis, colinear=colinear, **kwargs)

    def prepare_atoms(self, path, posfile, prefix_up, prefix_SOC, colinear=True):
        fname = os.path.join(path, posfile)
        try:
            print(f"Reading atomic structure from file {fname}.")
            atoms = read(os.path.join(path, posfile))
        except Exception:
            print(
                f"Cannot read atomic structure from file {fname}. Trying to read from Wannier input."
            )
            if colinear:
                fname = os.path.join(path, f"{prefix_up}.win")
            else:
                fname = os.path.join(path, f"{prefix_SOC}.win")

            print(f"Reading atomic structure from file {fname}.")
            atoms = parse_atoms(fname)
        return atoms

    def prepare_model_colinear(self, path, prefix_up, prefix_dn, atoms, output_path):
        print("Reading Wannier90 hamiltonian: spin up.")
        tbmodel_up = MyTB.read_from_wannier_dir(
            path=path, prefix=prefix_up, atoms=atoms, nls=False
        )
        print("Reading Wannier90 hamiltonian: spin down.")
        tbmodel_dn = MyTB.read_from_wannier_dir(
            path=path, prefix=prefix_dn, atoms=atoms, nls=False
        )
        basis, _ = auto_assign_basis_name(
            tbmodel_up.xred,
            atoms,
            write_basis_file=os.path.join(output_path, "assigned_basis.txt"),
        )
        tbmodels = (tbmodel_up, tbmodel_dn)
        return tbmodels, basis

    def prepare_model_ncl(self, path, prefix_SOC, atoms, groupby, output_path):
        print("Reading Wannier90 hamiltonian: non-colinear spin.")
        groupby = groupby.lower().strip()
        if groupby not in ["spin", "orbital"]:
            raise ValueError("groupby can only be spin or orbital.")
        tbmodel = MyTB.read_from_wannier_dir(
            path=path, prefix=prefix_SOC, atoms=atoms, groupby=groupby, nls=True
        )
        basis, _ = auto_assign_basis_name(
            tbmodel.xred,
            atoms,
            write_basis_file=os.path.join(output_path, "assigned_basis.txt"),
        )
        return tbmodel, basis

    def description(self, path, prefix_up, prefix_dn, prefix_SOC, colinear=True):
        if colinear:
            description = f""" Input from collinear Wannier90 data.
 Tight binding data from {path}.
 Prefix of wannier function files:{prefix_up} and {prefix_dn}.
Warning: Please check if the noise level of Wannier function Hamiltonian to make sure it is much smaller than the exchange values.
\n"""

        else:
            description = f""" Input from non-collinear Wannier90 data.
Tight binding data from {path}.
Prefix of wannier function files:{prefix_SOC}.
Warning: Please check if the noise level of Wannier function Hamiltonian to make sure it is much smaller than the exchange values.
The DMI component parallel to the spin orientation, the Jani which has the component of that orientation should be disregarded
e.g. if the spins are along z, the xz, yz, zz, zx, zy components and the z component of DMI.
If you need these component, try to do three calculations with spin along x, y, z,  or use structure with z rotated to x, y and z. And then use TB2J_merge.py to get the full set of parameters.\n"""

        return description


gen_exchange = WannierManager



import random
from dataclasses import dataclass
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
from numpy.linalg import matrix_rank
from scipy.interpolate import LinearNDInterpolator
from scipy.optimize import curve_fit


@dataclass
class Anisotropy:
    T: np.ndarray = None
    direction: np.ndarray = None
    amplitude: float = None
    isotropic_part: float = None

    @classmethod
    def from_T6(cls, T6):
        T = T6_to_T(T6)
        isotropic_part = np.trace(T) / 3
        T -= isotropic_part * np.eye(3)
        direction, amplitude = anisotropy_tensor_to_vector(T)
        return cls(
            direction=direction, amplitude=amplitude, isotropic_part=isotropic_part, T=T
        )

    @classmethod
    def from_direction_amplitude(cls, direction, amplitude, isotropic_part=0.0):
        T = anisotropy_vector_to_tensor(direction, amplitude)
        return cls(
            T=T, direction=direction, amplitude=amplitude, isotropic_part=isotropic_part
        )

    @classmethod
    def from_tensor(cls, T):
        isotropic_part = np.trace(T) / 3 * np.eye(3)
        T -= np.trace(T) / 3 * np.eye(3)
        direction, amplitude = anisotropy_tensor_to_vector(T)
        return cls(
            T=T, direction=direction, amplitude=amplitude, isotropic_part=isotropic_part
        )

    def __post_init__(self):
        if self.isotropic_part is None:
            self.isotropic_part = 0
        if self.T is None:
            self.T = (
                anisotropy_vector_to_tensor(self.direction, self.amplitude)
                + self.isotropic_part
            )
        elif self.direction is None or self.amplitude is None:
            self.isotropic_part = np.trace(self.T) / 3 * np.eye(3)
            # print(f'anisotropic tensor = {self.anisotropic_part}')
            self.direction, self.amplitude = anisotropy_tensor_to_vector(
                self.anisotropic_part
            )
            self.isotropic_part = np.trace(self.T) / 3
        if self.T is None or self.direction is None or self.amplitude is None:
            raise ValueError(
                "The input does not have enough information to create the anisotropy object."
            )

    def is_rank_one(self):
        """
        single-axis anisotropy should be rank-one.
        """
        print(f"rank = {matrix_rank(self.anisotropic_part)}")
        print(f"anisotropic_part = {self.anisotropic_part}")
        return matrix_rank(self.anisotropic_part) == 1

    def tensor(self):
        return self.T

    @property
    def anisotropic_part(self):
        return self.T - self.isotropic_part * np.eye(3)

    @property
    def axis(self):
        return self.direction

    @property
    def axis_type(self):
        if self.is_easy_axis():
            return "easy"
        else:
            return "hard"

    def axis_angle(self, unit="rad"):
        theta = np.arccos(self.direction[2])
        phi = np.arctan2(self.direction[1], self.direction[0])
        if unit.startswith("deg"):
            theta = theta * 180 / np.pi
            phi = phi * 180 / np.pi
        return theta, phi

    def amplitude(self):
        return self.amplitude

    def is_easy_axis(self):
        return self.amplitude > 0

    def is_hard_axis(self):
        return self.amplitude < 0

    def energy_vector_form(self, S=None, angle=None, include_isotropic=False):
        if S is None:
            S = sphere_to_cartesian(angle)
        print(f"S shape = {S.shape}")
        # return anisotropy_energy_vector_form(S, *angles, self.amplitude, self.isotropic_part)
        print(f"direction shape = {self.direction.shape}")
        print(f"amplitude shape = {self.amplitude.shape}")
        print(f"iso shape = {self.isotropic_part.shape}")
        E = -self.amplitude * (S @ self.direction) ** 2
        if include_isotropic:
            E = E + self.isotropic_part
        return E

    def energy_tensor_form(self, S=None, angle=None, include_isotropic=False):
        # return anisotropy_energy_tensor_form(self.T, S)
        if S is None:
            S = sphere_to_cartesian(angle)
        if include_isotropic:
            return -S.T @ self.T @ S
        else:
            return -S.T @ self.anisotropic_part @ S

    @classmethod
    def fit_from_data(cls, thetas, phis, values, test=False, units="rad"):
        """
        Fit the anisotropic tensor to the data
        parameters:
           thetas: the polar angle in degree
           phis: the azimuthal angle in degree
           values: the anisotropic value
        Return:
           the anisotropic object fitted from the data
        """
        angles = np.vstack([thetas, phis])
        if units.lower().startswith("deg"):
            angles = np.deg2rad(angles)
        params, cov = curve_fit(anisotropy_energy, angles, values)
        fitted_values = anisotropy_energy(angles, *params)

        delta = fitted_values - values

        # print(f'Max value = {np.max(values)}, Min value = {np.min(values)}')
        if np.abs(delta).max() > 1e-4:
            print("Warning: The fitting is not consistent with the data.")
            print(f"Max-min = {np.max(values) - np.min(values)}")
            print(f"delta = {np.max(np.abs(delta))}")
        T = T6_to_T(params)
        obj = cls(T=T)

        if test:
            values2 = []
            for i in range(len(thetas)):
                E = obj.energy_tensor_form(
                    angle=[thetas[i], phis[i]], include_isotropic=True
                )
                values2.append(E)
            # delta2 = np.array(values2) - values
            # print(delta2)

            ax = plot_3D_scatter(angles, values - np.min(values), color="r")
            plot_3D_scatter(angles, values2 - np.min(values), ax=ax, color="b")
            plt.show()

        return obj

    @classmethod
    def fit_from_data_vector_form(cls, thetas, phis, values):
        """
        Fit the anisotropic tensor to the data
        parameters:
           thetas: the polar angle in degree
           phis: the azimuthal angle in degree
           values: the anisotropic value
        Return:
           the anisotropic object fitted from the data
        """
        angles = np.vstack([thetas, phis])
        params, cov = curve_fit(anisotropy_energy_vector_form, angles, values)
        fitted_values = anisotropy_energy_vector_form(angles, *params)
        delta = fitted_values - values
        print(f"Max value = {np.max(values)}, Min value = {np.min(values)}")
        print(f"Max-min = {np.max(values) - np.min(values)}")
        print(f"delta = {delta}")
        theta_a, phi_a, amplitude, isotropic_part = params
        direction = sphere_to_cartesian([theta_a, phi_a])
        return cls.from_direction_amplitude(
            direction=direction, amplitude=amplitude, isotropic_part=isotropic_part
        )

    @classmethod
    def fit_from_data_file(cls, fname, method="tensor"):
        """
        Fit the anisotropic tensor to the data
        parameters:
           fname: the file name of the data
        Return:
           anisotropy: the anisotropic object
        """
        data = np.loadtxt(fname)
        theta, phi, value = data[:, 0], data[:, 1], data[:, 2]
        if method == "tensor":
            return cls.fit_from_data(theta, phi, value)
        elif method == "vector":
            return cls.fit_from_data_vector_form(theta, phi, value)
        else:
            raise ValueError(f"Unknown method {method}")

    @classmethod
    def fit_from_xyz_data_file(cls, fname, method="tensor"):
        """
        Fit the anisotropic tensor to the data with x y z val form
        parameters:
           fname: the file name of the data
        Return:
              anisotropy: the anisotropic object
        """
        data = np.loadtxt(fname)
        xyz, value = data[:, 0:3], data[:, 3]
        theta, phi = np.array([cartesian_to_sphere(t) for t in xyz]).T
        if method == "tensor":
            return cls.fit_from_data(theta, phi, value)
        elif method == "vector":
            return cls.fit_from_data_vector_form(theta, phi, value)
        else:
            raise ValueError(f"Unknown method {method}")

    def plot_3d(self, ax=None, figname=None, show=True, surface=True):
        """
        plot the anisotropic energy in all directions in 3D
        S is the spin unit vector
        """
        # theta, phi = np.meshgrid(theta, phi)
        # value = self.energy_tensor_form(sphere_to_cartesian([theta, phi]))
        # x, y, z = sphere_to_cartesian(theta, phi, )

        if surface:
            thetas = np.arange(0, 181, 1)
            phis = np.arange(0, 362, 1)

            X, Y = np.meshgrid(thetas, phis)
            Z = np.zeros(X.shape)
            for i in range(len(thetas)):
                for j in range(len(phis)):
                    # S = sphere_to_cartesian([thetas[i], phis[j]])
                    # E = self.energy_vector_form(angle=[thetas[i], phis[j]])
                    E = self.energy_tensor_form(angle=[thetas[i], phis[j]])
                    Z[j, i] = E
            Z = Z - np.min(Z)
            imax = np.argmax(Z)
            X_max, Y_max, Z_max = (
                X.flatten()[imax],
                Y.flatten()[imax],
                Z.flatten()[imax],
            )
            imin = np.argmin(Z)
            X_min, Y_min, Z_min = (
                X.flatten()[imin],
                Y.flatten()[imin],
                Z.flatten()[imin],
            )

            X, Y, Z = sphere_to_cartesian([X, Y], Z)
            X_max, Y_max, Z_max = sphere_to_cartesian([X_max, Y_max], r=Z_max)
            X_min, Y_min, Z_min = sphere_to_cartesian([X_min, Y_min], r=Z_min)
            fig = plt.figure()
            ax = fig.add_subplot(111, projection="3d")
            ax.plot_surface(X, Y, Z, cmap="viridis", alpha=0.5)
            # scatter the minimal and maximal points
            ax.scatter(X_max, Y_max, Z_max, color="r", marker="o")
            ax.scatter(X_min, Y_min, Z_min, color="b", marker="o")

            # draw the easy axis
            if self.is_easy_axis():
                color = "r"
            else:
                color = "b"
            d = self.direction
            d = d / np.sign(d[0])
            d *= np.abs(self.amplitude) * 2.5
            ax.quiver(
                -d[0],
                -d[1],
                -d[2],
                2 * d[0],
                2 * d[1],
                2 * d[2],
                color=color,
                arrow_length_ratio=0.03,
            )

        else:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection="3d")
            thetas = np.arange(0, 181, 5)
            phis = np.arange(0, 360, 5)
            Es = []
            angles = []
            for t in thetas:
                for p in phis:
                    # S = sphere_to_cartesian([t, p])
                    angles.append([t, p])
                    Es.append(
                        self.energy_tensor_form(angle=[t, p], include_isotropic=False)
                    )
                    # x, y, z = sphere_to_cartesian([t, p], r=E)
                    # ax.scatter(x,y, z, cmap='viridis')
            angles = np.array(angles)
            # plot_3D_scatter(angles.T, Es-np.min(Es), ax=None)
            plot_3D_scatter(angles.T, Es, ax=None)
        if figname is not None:
            plt.savefig(figname)
            plt.close()
        if show:
            plt.show()

    def plot_contourf(self, ax=None, figname=None, show=False):
        if ax is None:
            fig, ax = plt.subplots()
        X, Y = np.meshgrid(np.arange(0, 180, 1), np.arange(-180, 180, 1))
        Z = np.zeros(X.shape)
        ntheta, nphi = X.shape
        for i in range(ntheta):
            for j in range(nphi):
                E = self.energy_tensor_form(angle=[X[i, j], Y[i, j]])
                Z[i, j] = E
        # find the X, Y for min and max of Z
        X_max, Y_max = np.unravel_index(np.argmax(Z), Z.shape)
        X_min, Y_min = np.unravel_index(np.argmin(Z), Z.shape)
        X_max, Y_max = X[X_max, Y_max], Y[X_max, Y_max]
        X_min, Y_min = X[X_min, Y_min], Y[X_min, Y_min]
        c = ax.contourf(X, Y, Z, cmap="viridis", levels=200)
        # print(X_max, Y_max, X_min, Y_min)
        # ax.scatter(X_max, Y_max, color="r", marker="o")
        # ax.scatter(X_min, Y_min, color="b", marker="o")
        ax.set_xlabel("$\theta$ (degree)")
        ax.set_ylabel("$\phi$ degree")
        # ax.scatter(X_max, Y_max, color="r", marker="o")
        # ax.scatter(X_min, Y_min, color="r", marker="o")

        # colorbar
        _cbar = plt.colorbar(c, ax=ax)

        if figname is not None:
            plt.savefig(figname)
            plt.close()
        if show:
            print(f"Max = {X_max}, {Y_max}, Min = {X_min}, {Y_min}")
            plt.show()
        return ax

    def tensor_strings(self, include_isotropic=False, multiplier=1):
        """
        convert the energy tensor to strings for easy printing
        parameters:
        include_isotropic: if include the isotropic part
        multiplier: the multiplier for the tensor. Use for scaling the tensor by units.
        """
        if include_isotropic:
            T = self.T
        else:
            T = self.T
        strings = np.array2string(T * multiplier, precision=5, separator=" ")
        return strings


def plot_3D_scatter(angles, values, ax=None, **kwargs):
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")
    thetas, phis = angles
    for i in range(len(thetas)):
        E = values[i]
        t, p = thetas[i], phis[i]
        x, y, z = sphere_to_cartesian([t, p], r=E)
        ax.scatter(x, y, z, **kwargs)
    return ax


def plot_3D_surface(fname, figname=None, show=True):
    data = np.loadtxt(fname)
    theta, phi, value = data[:, 0], data[:, 1], data[:, 2]
    value = value - np.min(value)

    imax = np.argmax(value)
    angle_max = theta[imax], phi[imax]
    imin = np.argmin(value)
    angle_min = theta[imin], phi[imin]
    amplitude = np.max(value) - np.min(value)

    fig = plt.figure()
    ax = fig.add_subplot(111, projection="3d")
    x, y, z = sphere_to_cartesian([theta, phi], r=value)
    # ax.plot_surface(x, y, z, triangles=None, cmap='viridis')
    ax.scatter(x, y, z, cmap="viridis")
    # ax.scatter(theta,phi, value, cmap='viridis')
    # draw a surface plot with a color map.
    if figname is not None:
        plt.savefig(figname)
    if show:
        plt.show()
    plt.close()
    return angle_max, angle_min, amplitude


def plot_3D_surface_interpolation(fname, figname=None, show=True):
    data = np.loadtxt(fname)
    theta, phi, value = data[:, 0], data[:, 1], data[:, 2]

    imax = np.argmax(value)
    angle_max = theta[imax], phi[imax]
    imin = np.argmin(value)
    angle_min = theta[imin], phi[imin]
    amplitude = np.max(value) - np.min(value)

    value = value - np.min(value)
    # interploate the data
    interp = LinearNDInterpolator((theta, phi), value)
    thetas = np.arange(0, 181, 1)
    phis = np.arange(0, 340, 1)

    X, Y = np.meshgrid(thetas, phis)
    Z = interp(X, Y)
    print(Z)
    # print(np.max(Z), np.min(Z))
    Z = Z - np.min(Z)
    X, Y, Z = sphere_to_cartesian([X, Y], Z)
    fig = plt.figure()
    ax = fig.add_subplot(111, projection="3d")
    ax.plot_surface(X, Y, Z, cmap="viridis")
    if figname is not None:
        plt.savefig(figname)
    if show:
        plt.show()
    plt.close()
    return angle_max, angle_min, amplitude


def sphere_to_cartesian(angles, r=1):
    """
    Transform the spherical coordinates to the cartesian coordinates
    parameters:
    angles: the polar and azimuthal angle in degree
    r: the radius
    Return:
    x, y, z: the cartesian coordinates
    """
    # print(angles)
    theta, phi = angles
    theta = theta * np.pi / 180
    phi = phi * np.pi / 180
    x = r * np.sin(theta) * np.cos(phi)
    y = r * np.sin(theta) * np.sin(phi)
    z = r * np.cos(theta)
    return np.array([x, y, z])


def cartesian_to_sphere(xyz, unit="deg"):
    """
    Transform the cartesian coordinates to the spherical coordinates
    parameters:
    xyz: the cartesian coordinates
    Return:
    theta, phi: the polar and azimuthal angle in degree
    """
    x, y, z = xyz
    r = np.linalg.norm(xyz)
    theta = np.arccos(z / r)
    phi = np.arctan2(y, x)
    if unit.lower().startswith("deg"):
        theta = theta * 180 / np.pi
        phi = phi * 180 / np.pi
    elif unit.lower.startswith("rad"):
        pass
    else:
        raise ValueError("unit must be 'deg' or 'rad'")
    return np.array([theta, phi])


def anisotropy_energy(angle, Txx, Tyy, Tzz, Tyz, Tzx, Txy):
    """
    Calculate the anisotropic energy
    parameters:
    angle: the polar and azimuthal angle in degree
    Txx, Tyy, Tzz, Tyz, Tzx, Txy: the anisotropic tensor
    Return:
    E: the anisotropic energy
    """
    # transform the tensor to the matrix form
    T = np.array([[Txx, Txy, Tzx], [Txy, Tyy, Tyz], [Tzx, Tyz, Tzz]])

    S = sphere_to_cartesian(angle)
    E = -np.diag(S.T @ T @ S)
    return E


def anisotropy_energy_vector_form(S, k_theta, k_phi, amplitude, isotropic_part=0):
    """
    Calculate the anisotropic energy from the vector form
    parameters:
        S: the spin vector
        kx, ky, kz: the direction of the anisotropy
    """
    Scart = sphere_to_cartesian(S)
    k = sphere_to_cartesian([k_theta, k_phi])
    print(f"k shape = {k.shape}")
    print(f"Scart shape = {Scart.shape}")
    E = [-amplitude * (Si @ k) ** 2 + isotropic_part for Si in Scart.T]
    return E


def T6_to_T(T6):
    """
    Transform the anisotropic tensor to the matrix form
    """
    Txx, Tyy, Tzz, Tyz, Tzx, Txy = T6
    T = np.array([[Txx, Txy, Tzx], [Txy, Tyy, Tyz], [Tzx, Tyz, Tzz]])
    return T


def is_rank_one(T):
    """
    Check if the tensor is rank one
    """
    return matrix_rank(T) == 1


def fit_anisotropy(thetas, phis, values):
    """
    Fit the anisotropic tensor to the data
    parameters:
    theta: the polar angle in degree
    phi: the azimuthal angle in degree
    value: the anisotropic value
    Return:
     T: the anisotropic tensor
    """
    # transform the data to the cartesian coordinates
    # fit the data to the anisotropic energy
    angles = np.vstack([thetas, phis])
    params, cov = curve_fit(anisotropy_energy, angles, values)
    # check if the fitting is consistent with the data
    T = T6_to_T(params)
    direction, amp = anisotropy_tensor_to_vector(T)
    return direction, amp


def anisotropy_energy_vector_form2(direction, amplitude, S):
    """
    Calculate the anisotropic energy
    parameters:
    direction: the easy axis/ hard axis of the anisotropy
    amplitude: the amplitude of the anisotropy
    S: the spin vector
    Return:
    E: the anisotropic energy, E = - amplitude * (S @ direction)**2
    """
    # normalize the direction
    direction = direction / np.linalg.norm(direction)
    print(f"direction shape = {direction.shape}")
    E = -amplitude * (S @ direction) ** 2
    return E


def anisotropy_energy_tensor_form(T, S):
    """
    Calculate the anisotropic energy
    parameters:
    T: the anisotropic tensor
    S: the spin vector
    Return:
    E: the anisotropic energy, E = - S.T @ T @ S
    """
    E = -S.T @ T @ S
    return E


def anisotropy_vector_to_tensor(direction, amplitude):
    """
    Transform the anisotropic vector to the anisotropic tensor
    parameters:
    direction: the easy axis/ hard axis of the anisotropy (direction is normalized)
    amplitude: the amplitude of the anisotropy
    Return:
    T: the anisotropic tensor
    """
    direction = direction / np.linalg.norm(direction)
    T = amplitude * np.outer(direction, direction)
    return T


def anisotropy_tensor_to_vector(T):
    """
    Transform the anisotropic tensor to the anisotropic vector
    parameters:
    T: the anisotropic tensor
    Return:
    direction: the easy axis/ hard axis of the anisotropy
    amplitude: the amplitude of the anisotropy, if the anisotropy is positive, the easy axis is the easy axis, otherwise, the hard axis is the easy axis
    """
    w, v = np.linalg.eig(T)
    if not is_rank_one(T):
        # print("Warning: The anisotropy  tensor is not rank one. The tensor cannot be transformed to the vector form.")
        # print(f"The eigenvalues are {w}.")
        pass
    index = np.argmax(np.abs(w))
    direction = v[:, index]
    direction = direction / np.sign(direction[0])
    amplitude = w[index]
    return direction, amplitude


def test_anisotorpy_vector_to_tensor():
    direction = np.random.rand(3)
    direction = direction / np.linalg.norm(direction)
    amplitude = random.uniform(-1, 1)
    S = np.random.rand(3)
    S = S / np.linalg.norm(S)
    T = anisotropy_vector_to_tensor(direction, amplitude)
    E_vector = anisotropy_energy_vector_form(direction, amplitude, S)
    E_tensor = anisotropy_energy_tensor_form(T, S)
    diff = E_vector - E_tensor
    print(f"diff = {diff}")
    assert np.abs(diff) < 1e-10

    # test if the inverse transformation get the direction and amplitude back
    dir2, amp2 = anisotropy_tensor_to_vector(T)

    # set the first element of the direction to be positive
    if direction[0] * dir2[0] < 0:
        dir2 = -dir2
    diff = np.linalg.norm(dir2 - direction)
    # print(f'direction = {direction}, amplitude = {amplitude}')
    # print(f'dir2 = {dir2}, amp2 = {amp2}')
    assert diff < 1e-10
    assert np.abs(amp2 - amplitude) < 1e-10


def test_anisotropy_tensor_to_vector():
    T = np.random.rand(3, 3)
    T = T + T.T
    T = T - np.trace(T) / 3
    direction, amplitude = anisotropy_tensor_to_vector(T)
    T2 = anisotropy_vector_to_tensor(direction, amplitude)
    print(f"T = {T}")
    print(f"T2 = {T2}")
    diff = np.linalg.norm(T - T2)
    print(f"diff = {diff}")
    assert diff < 1e-10


def test_fit_anisotropy():
    data = np.loadtxt("anisotropy.dat")
    theta, phi, value = data[:, 0], data[:, 1], data[:, 2]
    angles = np.vstack([theta, phi])
    T6 = fit_anisotropy(theta, phi, value)
    T = T6_to_T(T6)
    if not is_rank_one(T):
        print("Warning: The anisotropy  tensor is not rank one. ")
    fitted_values = anisotropy_energy(angles, *T6)
    delta = fitted_values - value
    print(f"Max value = {np.max(value)}, Min value = {np.min(value)}")
    print(
        f"Max fitted value = {np.max(fitted_values)}, Min fitted value = {np.min(fitted_values)}"
    )
    for i in range(len(theta)):
        print(
            f"theta = {theta[i]}, phi = {phi[i]}, value = {value[i]}, fitted value = {fitted_values[i]}, delta = {delta[i]}"
        )


def view_anisotropy_strain():
    strains1 = [(x, 0.0) for x in np.arange(0.000, 0.021, 0.002)]
    strains2 = [(0.02, y) for y in np.arange(0.000, 0.021, 0.002)]
    strains = strains1 + strains2
    path = Path("anisotropy_strain_from_tensor")
    path.mkdir(exist_ok=True)
    fname = "a.dat"
    fh = open(fname, "w")
    fh.write("# s0  s1  axis   amplitude   axis_type angle_111 \n")
    for strain in strains:
        s0, s1 = strain
        ani = Anisotropy.fit_from_data_file(f"a{s0:.3f}_b{s1:.3f}_plusU.dat")
        dx, dy, dz = ani.direction
        angle_from_111 = (
            np.arccos(ani.direction @ np.array([1, 1, 1]) / np.sqrt(3)) * 180 / np.pi
        )
        fh.write(
            f"{s0:.3f}  {s1:.3f} ( {dx:.3f}  {dy:.3f}  {dz:.3f} ) {ani.amplitude:.5f} {ani.axis_type} {angle_from_111} \n"
        )
        ani.plot_3d(
            surface=True, figname=path / f"a{s0:.3f}_b{s1:.3f}_plusU.png", show=False
        )
    fh.close()


def view_anisotropy_strain_raw():
    strains1 = [(x, 0.0) for x in np.arange(0.000, 0.021, 0.002)]
    strains2 = [(0.02, y) for y in np.arange(0.000, 0.021, 0.002)]
    strains = strains1 + strains2
    path = Path("anisotropy_strain_raw")
    path.mkdir(exist_ok=True)
    fname = "a_raw.dat"
    fh = open(fname, "w")
    fh.write("# s0  s1  direction(max) direction(min)  amplitude\n")
    for strain in strains:
        print(f"strain = {strain}")
        s0, s1 = strain
        # plot_3D_surface(f"a{s0:.3f}_b{s1:.3f}_plusU.dat")
        angle_max, angle_min, amplitude = plot_3D_surface(
            f"a{s0:.3f}_b{s1:.3f}_plusU.dat",
            figname=path / f"a{s0:.3f}_b{s1:.3f}_plusU.png",
        )
        dmax = sphere_to_cartesian(angle_max)
        dmin = sphere_to_cartesian(angle_min)
        fh.write(
            f"{s0:.3f}  {s1:.3f} ( {dmax[0]:.3f}  {dmax[1]:.3f}  {dmax[2]:.3f} ) ({dmin[0]:.3f}  {dmin[1]:.3f}  {dmin[2]:.3f}) {amplitude:.5f} \n"
        )

        plt.close()


if __name__ == "__main__":
    # test_fit_anisotropy()
    # test_anisotorpy_vector_to_tensor()
    # test_anisotropy_tensor_to_vector()
    # ani = Anisotropy.fit_from_data("anisotropy.dat")
    # ani = Anisotropy.fit_from_data_file("a0.002_b0.000_plusU.dat")
    # ani = Anisotropy.fit_from_data_file("a0.002_b0.000_plusU.dat", method='tensor')
    # print(f'direction = {ani.direction}')
    # ani.plot_3d()
    # plot_3D_surface("a0.002_b0.000_plusU.dat")
    # plot_3D_surface_interpolation("a0.020_b0.020_plusU.dat")
    # plot_3D_surface("a0.020_b0.020_plusU.dat")
    # plot_3D_surface("a0.020_b0.000_plusU.dat")
    view_anisotropy_strain()
    # view_anisotropy_strain_raw()
    # s0=0.000
    # s1=0.000
    # plot_3D_surface_interpolation(f"a{s0:.3f}_b{s1:.3f}_plusU.dat", figname=None)



import copy
from pathlib import Path

import numpy as np
from sympair import SymmetryPairFinder

from TB2J.io_exchange import SpinIO
from TB2J.versioninfo import print_license


class TB2JSymmetrizer:
    def __init__(self, exc, symprec=1e-8, verbose=True, Jonly=False):
        # list of pairs with the index of atoms
        ijRs = exc.ijR_list_index_atom()
        finder = SymmetryPairFinder(atoms=exc.atoms, pairs=ijRs, symprec=symprec)
        self.verbose = verbose
        self.Jonly = Jonly

        if verbose:
            print("=" * 30)
            print_license()
            print("-" * 30)
            print(
                "WARNING: The symmetry detection is based on the crystal symmetry, not the magnetic symmetry. Make sure if this is what you want."
            )
            print("-" * 30)
            if exc.has_dmi:
                print(
                    "WARNING: Currently only the isotropic exchange is symmetrized. Symmetrization of DMI and anisotropic exchange are not yet implemented."
                )

            print(f"Finding crystal symmetry with symprec of {symprec} Angstrom.")
            print("Symmetry found:")
            print(finder.spacegroup)
            print("-" * 30)
        self.pgdict = finder.get_symmetry_pair_group_dict()
        self.exc = exc
        self.new_exc = copy.deepcopy(exc)
        self.Jonly = Jonly

    def print_license(self):
        print_license()

    def symmetrize_J(self):
        """
        Symmetrize the exchange parameters J.
        """
        symJdict = {}
        # Jdict = self.exc.exchange_Jdict
        # ngroup = self.pgdict
        for pairgroup in self.pgdict.groups:
            ijRs = pairgroup.get_all_ijR()
            ijRs_spin = [self.exc.ijR_index_atom_to_spin(*ijR) for ijR in ijRs]
            Js = [self.exc.get_J(*ijR_spin) for ijR_spin in ijRs_spin]
            Javg = np.average(Js)
            for i, j, R in ijRs_spin:
                symJdict[(R, i, j)] = Javg
        self.new_exc.exchange_Jdict = symJdict
        if self.Jonly:
            self.new_exc.has_dmi = False
            self.new_exc.dmi_ddict = None
            self.new_exc.has_bilinear = False
            self.new_exc.Jani_dict = None
            self.has_uniaxial_anisotropy = False
            self.k1 = None
            self.k1dir = None

    def output(self, path="TB2J_symmetrized"):
        if path is None:
            path = Path(".")
        self.new_exc.write_all(path=path)

    def run(self, path=None):
        print("** Symmetrizing exchange parameters.")
        self.symmetrize_J()
        print("** Outputing the symmetrized exchange parameters.")
        print(f"** Output path: {path} .")
        self.output(path=path)
        print("** Finished.")


def symmetrize_J(
    exc=None,
    path=None,
    fname="TB2J.pickle",
    symprec=1e-5,
    output_path="TB2J_symmetrized",
    Jonly=False,
):
    """
    symmetrize the exchange parameters
    parameters:
    exc: exchange
    """
    if exc is None:
        exc = SpinIO.load_pickle(path=path, fname=fname)
    symmetrizer = TB2JSymmetrizer(exc, symprec=symprec, Jonly=Jonly)
    symmetrizer.run(path=output_path)


def symmetrize_J_cli():
    from argparse import ArgumentParser

    parser = ArgumentParser(
        description="Symmetrize exchange parameters. Currently, it take the crystal symmetry into account and  not the magnetic moment."
    )
    parser.add_argument(
        "-i",
        "--inpath",
        default=None,
        help="input path to the exchange parameters",
    )
    parser.add_argument(
        "-o",
        "--outpath",
        default="TB2J_results_symmetrized",
        help="output path to the symmetrized exchange parameters",
    )
    parser.add_argument(
        "-s",
        "--symprec",
        type=float,
        default=1e-5,
        help="precision for symmetry detection. default is 1e-5 Angstrom",
    )

    parser.add_argument(
        "--Jonly",
        action="store_true",
        help="symmetrize only the exchange parameters and discard the DMI and anisotropic exchange",
        default=False,
    )

    args = parser.parse_args()
    if args.inpath is None:
        parser.print_help()
        raise ValueError("Please provide the input path to the exchange.")
    symmetrize_J(
        path=args.inpath,
        output_path=args.outpath,
        symprec=args.symprec,
        Jonly=args.Jonly,
    )


if __name__ == "__main__":
    symmetrize_J_cli()



#!/usr/bin/env python3
import argparse
import sys

from TB2J.exchange_params import add_exchange_args_to_parser
from TB2J.interfaces import gen_exchange_abacus
from TB2J.versioninfo import print_license


def run_abacus2J():
    print_license()
    parser = argparse.ArgumentParser(
        description="abacus2J: Using magnetic force theorem to calculate exchange parameter J from abacus Hamiltonian in the LCAO mode"
    )
    # Add ABACUS specific arguments
    parser.add_argument(
        "--path", help="the path of the abacus calculation", default="./", type=str
    )
    parser.add_argument(
        "--suffix",
        help="the label of the abacus calculation. There should be an output directory called OUT.suffix",
        default="abacus",
        type=str,
    )

    # Add common exchange arguments
    parser = add_exchange_args_to_parser(parser)

    args = parser.parse_args()

    index_magnetic_atoms = args.index_magnetic_atoms
    if index_magnetic_atoms is not None:
        index_magnetic_atoms = [i - 1 for i in index_magnetic_atoms]

    if args.elements is None and index_magnetic_atoms is None:
        print("Please input the magnetic elements, e.g. --elements Fe Ni")
        sys.exit()

    # include_orbs = {}

    gen_exchange_abacus(
        path=args.path,
        suffix=args.suffix,
        kmesh=args.kmesh,
        magnetic_elements=args.elements,
        include_orbs={},
        Rcut=args.rcut,
        emin=args.emin,
        nz=args.nz,
        description=args.description,
        output_path=args.output_path,
        use_cache=args.use_cache,
        nproc=args.np,
        exclude_orbs=args.exclude_orbs,
        orb_decomposition=args.orb_decomposition,
        index_magnetic_atoms=index_magnetic_atoms,
    )


if __name__ == "__main__":
    run_abacus2J()



import numpy as np


def theta_phi_even_spaced(n):
    """
    Return n evenly spaced theta and phi values in the ranges [0, pi] and [0, 2*pi] respectively.
    """
    phis = []
    thetas = []
    for i in range(n):
        phi = 2 * np.pi * i / n
        phis.append(phi)
        r = np.sin(np.pi * i / n)
        theta = np.arccos(1 - 2 * r)
        thetas.append(theta)
    return thetas, phis



#!/usr/bin/env python3
import argparse
import os
import sys
from TB2J.io_merge import merge


def main():
    parser = argparse.ArgumentParser(
        description="TB2J_merge: merge the exchange parameters calculated from different directions. It can accept two types of calculations, first, the lattice structures are rotated from the original structure, from z to x, y, and z, respectively. The structures could be generated using the TB2J_rotate.py command. The second type is the three calculations have the spins along the x, y, and z axis, respectively."
    )
    parser.add_argument(
        "directories",
        type=str,
        nargs="+",
        help="The three directories the TB2J calculations are done. Inside each of them, there should be a TB2J_results directory which contains the magnetic interaction parameter files. e.g. Fe_x Fe_y Fe_z. Alternatively, it could  be the TB2J results directories.",
    )

    parser.add_argument(
        "--type",
        "-T",
        type=str,
        help="The type of calculations, either structure of spin, meaning that the three calculations are done by rotating the structure/spin. ",
    )
    parser.add_argument(
        "--output_path",
        help="The path of the output directory, default is TB2J_results",
        type=str,
        default="TB2J_results",
    )
    parser.add_argument(
        "--main_path",
        help="The path containning the reference structure.",
        type=str,
        default=None
    )

    args = parser.parse_args()
    # merge(*(args.directories), args.type.strip().lower(), path=args.output_path)
    # merge(*(args.directories), method=args.type.strip().lower(), path=args.output_path)
    #merge2(args.directories, args.type.strip().lower(), path=args.output_path)
    merge(*args.directories, main_path=args.main_path, write_path=args.output_path)


main()



import re
from collections import defaultdict

import numpy as np


def split_orb_name(name):
    """
    split name to : n, l, label
    """
    m = re.findall(r"([a-z\d\-\^\*]*)(.*)", name)
    m = m[0]
    return m[0], m[1]


def map_orbs_matrix(orblist, spinor=False, include_only=None):
    """
    map the orbitals to a matrix
    Method:
    1. split the orbital name to n, l, label
    2. group the orbitals by n, l
    3. create a matrix with 1 for each orbital in the group
    4. return the matrix and the group names
    """

    if spinor:
        orblist = orblist[::2]

    norb = len(orblist)

    # print("orblist: ", orblist)
    ss = [split_orb_name(orb) for orb in orblist]
    orbdict = dict(zip(ss, range(norb)))

    reduced_orbdict = defaultdict(lambda: [])

    # print(f"Orbital dictionary: {orbdict}")
    # print("include_only: ", include_only)

    if include_only is None:
        for key, val in orbdict.items():
            reduced_orbdict[key[0]].append(val)
    else:
        for key, val in orbdict.items():
            if key[0][:2] in include_only or key[0][:1] in include_only:
                # [:2] for 3d, 4d, 5d, etc. and [:1] for s, p, d, etc
                reduced_orbdict[key[0]].append(val)

    # print(f"reduced_orbdict: {reduced_orbdict}")
    reduced_orbs = tuple(reduced_orbdict.keys())
    ngroup = len(reduced_orbdict)
    mmat = np.zeros((norb, ngroup), dtype=int)

    for i, (key, val) in enumerate(reduced_orbdict.items()):
        for j in val:
            mmat[j, i] = 1
    return mmat, reduced_orbs


def test_split():
    split_orb_name("3sZ1")
    split_orb_name("3dxyZ1")
    split_orb_name("5dxyZ1")
    split_orb_name("5dx2-y2Z1P")


def test():
    odict = {
        0: [
            "3sZ1",
            "3sZ1",
            "4sZ1",
            "4sZ1",
            "4sZ2",
            "4sZ2",
            "3pyZ1",
            "3pyZ1",
            "3pzZ1",
            "3pzZ1",
            "3pxZ1",
            "3pxZ1",
            "3dxyZ1",
            "3dxyZ1",
            "3dyzZ1",
            "3dyzZ1",
            "3dz2Z1",
            "3dz2Z1",
            "3dxzZ1",
            "3dxzZ1",
            "3dx2-y2Z1",
            "3dx2-y2Z1",
            "3dxyZ2",
            "3dxyZ2",
            "3dyzZ2",
            "3dyzZ2",
            "3dz2Z2",
            "3dz2Z2",
            "3dxzZ2",
            "3dxzZ2",
            "3dx2-y2Z2",
            "3dx2-y2Z2",
            "4pyZ1P",
            "4pyZ1P",
            "4pzZ1P",
            "4pzZ1P",
            "4pxZ1P",
            "4pxZ1P",
        ],
        1: [
            "3sZ1",
            "3sZ1",
            "4sZ1",
            "4sZ1",
            "4sZ2",
            "4sZ2",
            "3pyZ1",
            "3pyZ1",
            "3pzZ1",
            "3pzZ1",
            "3pxZ1",
            "3pxZ1",
            "3dxyZ1",
            "3dxyZ1",
            "3dyzZ1",
            "3dyzZ1",
            "3dz2Z1",
            "3dz2Z1",
            "3dxzZ1",
            "3dxzZ1",
            "3dx2-y2Z1",
            "3dx2-y2Z1",
            "3dxyZ2",
            "3dxyZ2",
            "3dyzZ2",
            "3dyzZ2",
            "3dz2Z2",
            "3dz2Z2",
            "3dxzZ2",
            "3dxzZ2",
            "3dx2-y2Z2",
            "3dx2-y2Z2",
            "4pyZ1P",
            "4pyZ1P",
            "4pzZ1P",
            "4pzZ1P",
            "4pxZ1P",
            "4pxZ1P",
        ],
        2: [
            "5sZ1",
            "5sZ1",
            "5sZ2",
            "5sZ2",
            "5pyZ1",
            "5pyZ1",
            "5pzZ1",
            "5pzZ1",
            "5pxZ1",
            "5pxZ1",
            "5pyZ2",
            "5pyZ2",
            "5pzZ2",
            "5pzZ2",
            "5pxZ2",
            "5pxZ2",
            "5dxyZ1P",
            "5dxyZ1P",
            "5dyzZ1P",
            "5dyzZ1P",
            "5dz2Z1P",
            "5dz2Z1P",
            "5dxzZ1P",
            "5dxzZ1P",
            "5dx2-y2Z1P",
            "5dx2-y2Z1P",
        ],
        3: [
            "5sZ1",
            "5sZ1",
            "5sZ2",
            "5sZ2",
            "5pyZ1",
            "5pyZ1",
            "5pzZ1",
            "5pzZ1",
            "5pxZ1",
            "5pxZ1",
            "5pyZ2",
            "5pyZ2",
            "5pzZ2",
            "5pzZ2",
            "5pxZ2",
            "5pxZ2",
            "5dxyZ1P",
            "5dxyZ1P",
            "5dyzZ1P",
            "5dyzZ1P",
            "5dz2Z1P",
            "5dz2Z1P",
            "5dxzZ1P",
            "5dxzZ1P",
            "5dx2-y2Z1P",
            "5dx2-y2Z1P",
        ],
        4: [
            "5sZ1",
            "5sZ1",
            "5sZ2",
            "5sZ2",
            "5pyZ1",
            "5pyZ1",
            "5pzZ1",
            "5pzZ1",
            "5pxZ1",
            "5pxZ1",
            "5pyZ2",
            "5pyZ2",
            "5pzZ2",
            "5pzZ2",
            "5pxZ2",
            "5pxZ2",
            "5dxyZ1P",
            "5dxyZ1P",
            "5dyzZ1P",
            "5dyzZ1P",
            "5dz2Z1P",
            "5dz2Z1P",
            "5dxzZ1P",
            "5dxzZ1P",
            "5dx2-y2Z1P",
            "5dx2-y2Z1P",
        ],
        5: [
            "5sZ1",
            "5sZ1",
            "5sZ2",
            "5sZ2",
            "5pyZ1",
            "5pyZ1",
            "5pzZ1",
            "5pzZ1",
            "5pxZ1",
            "5pxZ1",
            "5pyZ2",
            "5pyZ2",
            "5pzZ2",
            "5pzZ2",
            "5pxZ2",
            "5pxZ2",
            "5dxyZ1P",
            "5dxyZ1P",
            "5dyzZ1P",
            "5dyzZ1P",
            "5dz2Z1P",
            "5dz2Z1P",
            "5dxzZ1P",
            "5dxzZ1P",
            "5dx2-y2Z1P",
            "5dx2-y2Z1P",
        ],
        6: [
            "5sZ1",
            "5sZ1",
            "5sZ2",
            "5sZ2",
            "5pyZ1",
            "5pyZ1",
            "5pzZ1",
            "5pzZ1",
            "5pxZ1",
            "5pxZ1",
            "5pyZ2",
            "5pyZ2",
            "5pzZ2",
            "5pzZ2",
            "5pxZ2",
            "5pxZ2",
            "5dxyZ1P",
            "5dxyZ1P",
            "5dyzZ1P",
            "5dyzZ1P",
            "5dz2Z1P",
            "5dz2Z1P",
            "5dxzZ1P",
            "5dxzZ1P",
            "5dx2-y2Z1P",
            "5dx2-y2Z1P",
        ],
        7: [
            "5sZ1",
            "5sZ1",
            "5sZ2",
            "5sZ2",
            "5pyZ1",
            "5pyZ1",
            "5pzZ1",
            "5pzZ1",
            "5pxZ1",
            "5pxZ1",
            "5pyZ2",
            "5pyZ2",
            "5pzZ2",
            "5pzZ2",
            "5pxZ2",
            "5pxZ2",
            "5dxyZ1P",
            "5dxyZ1P",
            "5dyzZ1P",
            "5dyzZ1P",
            "5dz2Z1P",
            "5dz2Z1P",
            "5dxzZ1P",
            "5dxzZ1P",
            "5dx2-y2Z1P",
            "5dx2-y2Z1P",
        ],
    }

    olist = odict[0]
    r1 = map_orbs_matrix(olist, spinor=True)
    print(r1)
    r2 = map_orbs_matrix(olist, spinor=True, include_only=["3d"])
    print(r2)


if __name__ == "__main__":
    test()



"""
This file is copied from the p_tqdm package. It is copied because the pip version, which is not compatible with python3.10, is not update to date as of the day of 30th, April, 2022. Once it is updated, we'll switch back to the p_tqdm library on pip. 

Below is the original licence:
==================================================
Copyright (c) 2020 Kyle Swanson                                                       

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
"""Map functions with tqdm progress bars for parallel and sequential processing.

p_map: Performs a parallel ordered map.
p_imap: Returns an iterator for a parallel ordered map.
p_umap: Performs a parallel unordered map.
p_uimap: Returns an iterator for a parallel unordered map.
t_map: Performs a sequential map.
t_imap: Returns an iterator for a sequential map.
"""

from collections.abc import Sized
from typing import Any, Callable, Generator, Iterable, List

from pathos.helpers import cpu_count
from pathos.multiprocessing import ProcessPool as Pool
from tqdm.auto import tqdm


def _parallel(
    ordered: bool, function: Callable, *iterables: Iterable, **kwargs: Any
) -> Generator:
    """Returns a generator for a parallel map with a progress bar.

    Arguments:
        ordered(bool): True for an ordered map, false for an unordered map.
        function(Callable): The function to apply to each element of the given Iterables.
        iterables(Tuple[Iterable]): One or more Iterables containing the data to be mapped.

    Returns:
        A generator which will apply the function to each element of the given Iterables
        in parallel in order with a progress bar.
    """

    # Extract num_cpus
    num_cpus = kwargs.pop("num_cpus", None)

    # Determine num_cpus
    if num_cpus is None:
        num_cpus = cpu_count()
    elif type(num_cpus) == float:
        num_cpus = int(round(num_cpus * cpu_count()))

    # Determine length of tqdm (equal to length of shortest iterable or total kwarg), if possible
    total = kwargs.pop("total", None)
    lengths = [len(iterable) for iterable in iterables if isinstance(iterable, Sized)]
    length = total or (min(lengths) if lengths else None)

    # Create parallel generator
    map_type = "imap" if ordered else "uimap"
    pool = Pool(num_cpus)
    map_func = getattr(pool, map_type)

    for item in tqdm(map_func(function, *iterables), total=length, **kwargs):
        yield item

    pool.clear()


def p_map(function: Callable, *iterables: Iterable, **kwargs: Any) -> List[Any]:
    """Performs a parallel ordered map with a progress bar."""

    ordered = True
    generator = _parallel(ordered, function, *iterables, **kwargs)
    result = list(generator)

    return result


def p_imap(function: Callable, *iterables: Iterable, **kwargs: Any) -> Generator:
    """Returns a generator for a parallel ordered map with a progress bar."""

    ordered = True
    generator = _parallel(ordered, function, *iterables, **kwargs)

    return generator


def p_umap(function: Callable, *iterables: Iterable, **kwargs: Any) -> List[Any]:
    """Performs a parallel unordered map with a progress bar."""

    ordered = False
    generator = _parallel(ordered, function, *iterables, **kwargs)
    result = list(generator)

    return result


def p_uimap(function: Callable, *iterables: Iterable, **kwargs: Any) -> Generator:
    """Returns a generator for a parallel unordered map with a progress bar."""

    ordered = False
    generator = _parallel(ordered, function, *iterables, **kwargs)

    return generator


def _sequential(function: Callable, *iterables: Iterable, **kwargs: Any) -> Generator:
    """Returns a generator for a sequential map with a progress bar.

    Arguments:
        function(Callable): The function to apply to each element of the given Iterables.
        iterables(Tuple[Iterable]): One or more Iterables containing the data to be mapped.

    Returns:
        A generator which will apply the function to each element of the given Iterables
        sequentially in order with a progress bar.
    """

    # Determine length of tqdm (equal to length of shortest iterable)
    length = min(len(iterable) for iterable in iterables if isinstance(iterable, Sized))

    # Create sequential generator
    for item in tqdm(map(function, *iterables), total=length, **kwargs):
        yield item


def t_map(function: Callable, *iterables: Iterable, **kwargs: Any) -> List[Any]:
    """Performs a sequential map with a progress bar."""

    generator = _sequential(function, *iterables, **kwargs)
    result = list(generator)

    return result


def t_imap(function: Callable, *iterables: Iterable, **kwargs: Any) -> Generator:
    """Returns a generator for a sequential map with a progress bar."""

    generator = _sequential(function, *iterables, **kwargs)

    return generator



try:
    from gpaw.lcao.tightbinding import TightBinding
    from gpaw import restart, GPAW
    from gpaw.lcao.tools import (
        get_lead_lcao_hamiltonian,
        get_lcao_hamiltonian,
        get_bf_centers,
        get_bfi,
    )

    _has_gpaw = True
except:
    _has_gpaw = False

# from banddownfolder.scdm.downfolder import BandDownfolder
import numpy as np
from scipy.linalg import eigh
from ase.dft.kpoints import monkhorst_pack
from ase.units import Ha
import pickle


class GPAWWrapper:
    """
    https://gitlab.com/gpaw/gpaw/-/blob/master/gpaw/lcao/tightbinding.py
    """

    def __init__(self, atoms=None, calc=None, gpw_fname=None):
        if not _has_gpaw:
            raise ImportError("GPAW was not imported. Please install gpaw.")
        if atoms is None or calc is None:
            atoms, calc = restart(gpw_fname, fixdensity=False, txt="nscf.txt")
        self.atoms = atoms
        self.calc = calc
        # self.wfs=self.calc.wfs
        # self.h=self.calc.hamiltonian
        # self.kpt_u = self.wfs.kpt_u
        self.cell = self.atoms.get_cell()
        # self.wann_centers=get_bf_centers(self.atoms)
        self.positions = self.cell.scaled_positions(get_bf_centers(self.atoms))
        # if calc.get_spin_polarized():
        #    self.positions=np.vstack([self.positions, self.positions])
        self.efermi = self.calc.get_fermi_level()
        self.R2kfactor = -2.0j * np.pi
        # self.nbasis=len(self.positions)
        self.norb = len(self.positions)
        self.nbasis = self.norb * 2
        self._name = "GPAW"

    # def gen_ham(self, k):
    #    H_MM=self.wfs.eigensolver.calculate_hamiltonian_matrix(self.h, self.wfs, k)
    #    tri2full(H_MM*Ha)
    #    return H_MM

    # def Sk(self, k):
    #    S_MM=self.wfs.S_qMM[kpt.q]
    #    tri2full(S_MM)
    #    return S_MM

    # def solve(self, k):
    #    return eigh(self.hamk(k), self.S(k))

    def get_kpts(self):
        return self.calc.get_ibz_k_points()

    def HS_and_eigen(self, kpts=None, convention=2):
        if kpts is not None:
            self.calc.set(kpts=kpts)
            self.atoms.calc = self.calc
            self.atoms.get_total_energy()

        self.calc.set(symmetry="off", fixdensity=True)
        self.atoms.get_total_energy()
        # self.kpt_u = self.wfs.kpt_u
        H, S = get_lcao_hamiltonian(self.calc)
        np.save("kpts.npy", kpts)
        np.save("H.npy", H)
        np.save("S.npy", S)

        nspin, nkpt, norb, _ = H.shape
        nbasis = nspin * norb
        evals = np.zeros((nkpt, nbasis), dtype=float)
        evecs = np.zeros((nkpt, nbasis, nbasis), dtype=complex)
        if self.calc.get_spin_polarized():
            H2 = np.zeros((nkpt, nbasis, nbasis), dtype=complex)
            # spin up
            H2[:, :norb, :norb] = H[0]
            # spin down
            H2[:, norb:, norb:] = H[1]
            S2 = np.zeros((nkpt, nbasis, nbasis), dtype=complex)
            S2[:, :norb, :norb] = S
            S2[:, norb:, norb:] = S

            for ikpt, k in enumerate(self.calc.get_ibz_k_points()):
                evals0, evecs0 = eigh(H[0, ikpt, :, :], S[ikpt, :, :])
                evals1, evecs1 = eigh(H[1, ikpt, :, :], S[ikpt, :, :])
                evals[ikpt, :norb] = evals0
                evals[ikpt, norb:] = evals1
                evecs[ikpt, :norb, :norb] = evecs0
                evecs[ikpt, norb:, norb:] = evecs1
        else:
            H2 = H[0]
            for ikpt, k in self.kpt_u:
                evals[ikpt], evecs[ikpt] = eigh(H[0, ikpt], S[ikpt])
        np.save("kpts.npy", kpts)
        np.save("H2.npy", H2)
        np.save("S2.npy", S2)
        np.save("evals.npy", evals)
        np.save("evecs.npy", evecs)
        return H2, S2, evals, evecs

    def solve_all(self, kpts, convention=2):
        H, evals, evecs = self.H_and_eigen(kpts=kpts, convention=convention)
        return evals, evecs


class GPAWTBWrapper:
    def __init__(self, calc=None, atoms=None, gpw_fname=None, pickle_fname=None):
        if not _has_gpaw:
            raise ImportError("GPAW was not imported. Please install gpaw.")
        if pickle_fname is None:
            if gpw_fname is not None:
                calc = GPAW(
                    gpw_fname,
                    fixdensity=True,
                    symmetry="off",
                    txt="TB2J_wrapper.log",
                    basis="dzp",
                    mode="lcao",
                )
                atoms = calc.atoms
                atoms.calc = calc
                atoms.get_potential_energy()
            tb = TightBinding(atoms, calc)
            self.H_NMM, self.S_NMM = tb.h_and_s()
            self.Rlist = tb.R_cN.T
        else:
            with open(gpw_fname, "rb") as myfile:
                self.H_NMM, self.S_NMM, self.Rlist = pickle.load(myfile)
        self.nR, self.nbasis, _ = self.H_NMM.shape
        self.positions = np.zeros((self.nbasis, 3))

    def save_pickle(self, fname):
        with open(fname, "wb") as myfile:
            pickle.dump([self.H_NMM, self.S_NMM, self.Rlist], myfile)

    def solve(self, k):
        self.Hk = np.zeros((self.nbasis, self.nbasis), dtype="complex")
        self.Sk = np.zeros((self.nbasis, self.nbasis), dtype="complex")
        for iR, R in enumerate(self.Rlist):
            phase = np.exp(-2j * np.pi * np.dot(k, R))
            self.Hk += self.H_NMM[iR] * phase
            self.Sk += self.S_NMM[iR] * phase
        self.Sk = np.real((self.Sk + self.Sk.T) / 2)
        return eigh(self.Hk, self.Sk)
        try:
            return eigh(self.Hk, self.Sk)
        except:
            return np.zeros(self.nbasis), np.zeros((self.nbasis, self.nbasis))

    def solve_all(self, kpts):
        evals = []
        evecs = []
        for ik, k in enumerate(kpts):
            evalue, evec = self.solve(k)
            evals.append(evalue)
            evecs.append(evec)
        return np.array(evals, dtype=float), np.array(evecs, dtype=complex, order="C")


def test_Ham():
    calc = GPAW(
        "/home/hexu/projects/gpaw/bccFe/atoms_nscf.gpw",
        fixdensity=True,
        symmetry="off",
        txt="nscf.txt",
    )
    atoms = calc.atoms
    atoms.calc = calc
    calc.set(kpts=(3, 3, 3))
    atoms.get_potential_energy()
    tb = TightBinding(atoms, calc)
    kpath = monkhorst_pack([3, 3, 3])
    evals, evecs = tb.band_structure(kpath, blochstates=True)


def test():
    g = GPAWWrapper(fname="STO_nscf.gpw")
    g.H_and_eigen(kpts=monkhorst_pack(2, 2, 2))
    # g.save_pickle('model.pickle')
    # df=GPAWDownfolder('model.pickle')
    # df.downfold(nwann=1, anchor_kpt=[0,0,0] )
    # df.plot_band_fitting(npoints=100)


# test_Ham()


def test():
    g = GPAWWrapper(gpw_fname="/home/hexu/projects/gpaw/bccFe/atoms_nscf.gpw")
    H, E, V = g.H_and_eigen(kpts=monkhorst_pack([2, 2, 2]))
    print(H.shape)
    print(E.shape)
    print(E)
    print(V.shape)



import os
from collections import defaultdict

import numpy as np
from ase.dft.kpoints import monkhorst_pack

from TB2J.io_exchange import SpinIO
from TB2J.Jtensor import decompose_J_tensor


def ind_to_indn(ind, n=3):
    """index to index. e.g. 1, 2, 4 to 1*3, 1*3+1, 1*3+2, ... 4*3, 4*3+1, 4*3+2
    :param ind: the input indices
    :returns: ind3, the output indices
    """
    indn = np.repeat(ind, n) * n
    for i in range(n):
        indn[i::n] += i
    return indn


class JR_model:
    def __init__(self, JR, Rlist):
        self.JR = JR
        self.Rlist = Rlist
        self.nR = len(Rlist)

    def get_Jq(self, q):
        Jq = np.zeros(self.JR[0].shape, dtype=complex)
        for iR, R in enumerate(self.Rlist):
            phase = np.exp(2.0j * np.pi * np.dot(q, R))
            Jq += self.JR[iR] * phase
        return Jq


class JDownfolder:
    def __init__(self, JR, Rlist, iM, iL, qmesh, iso_only=False):
        self.model = JR_model(JR, Rlist)
        self.Rlist = Rlist
        self.nR = len(Rlist)
        self.nM = len(iM)
        self.nL = len(iL)
        self.nsite = self.nM + self.nL
        self.iM = iM
        self.iL = iL
        self.qmesh = qmesh
        self.qpts = monkhorst_pack(qmesh)
        self.nqpt = len(self.qpts)
        self.nMn = self.nM * 3
        self.nLn = self.nL * 3
        self.iso_only = iso_only

    def get_JR(self):
        JR_downfolded = np.zeros((self.nR, self.nMn, self.nMn), dtype=float)
        Jq_downfolded = np.zeros((self.nqpt, self.nMn, self.nMn), dtype=complex)
        self.iMn = ind_to_indn(self.iM, n=3)
        self.iLn = ind_to_indn(self.iL, n=3)
        for iq, q in enumerate(self.qpts):
            Jq = self.model.get_Jq(q)
            Jq_downfolded[iq] = self.downfold_oneq(Jq)
            for iR, R in enumerate(self.Rlist):
                phase = np.exp(-2.0j * np.pi * np.dot(q, R))
                JR_downfolded[iR] += np.real(Jq_downfolded[iq] * phase / self.nqpt)
        return JR_downfolded, self.Rlist

    def downfold_oneq(self, J):
        JMM = J[np.ix_(self.iMn, self.iMn)]
        JLL = J[np.ix_(self.iLn, self.iLn)]
        JLM = J[np.ix_(self.iLn, self.iMn)]
        JML = J[np.ix_(self.iMn, self.iLn)]
        Jn = JMM - JML @ np.linalg.inv(JLL) @ JLM
        return Jn


class PWFDownfolder:
    def __init__(self, JR, Rlist, iM, iL, qmesh, atoms=None, iso_only=False, **kwargs):
        from lawaf.interfaces.magnon.magnon_downfolder import (
            MagnonDownfolder,
            MagnonWrapper,
        )

        model = MagnonWrapper(JR, Rlist, atoms)
        wann = MagnonDownfolder(model)
        # Downfold the band structure.
        index_basis = []
        for i in iM:
            index_basis += list(range(i * 3, i * 3 + 3))
        params = dict(
            method="projected",
            # method="maxprojected",
            kmesh=qmesh,
            nwann=len(index_basis),
            selected_basis=index_basis,
            # anchors={(0, 0, 0): (-1, -2, -3, -4)},
            # anchors={(0, 0, 0): ()},
            # use_proj=True,
            enhance_Amn=0.0,
        )
        params.update(kwargs)
        wann.set_parameters(**params)
        print("begin downfold")
        ewf = wann.downfold()
        # ewf.save_hr_pickle("downfolded_JR.pickle")

        # Plot the band structure.
        wann.plot_band_fitting(
            # kvectors=np.array([[0, 0, 0], [0.5, 0, 0],
            #                   [0.5, 0.5, 0], [0, 0, 0],
            #                   [.5, .5, .5]]),
            # knames=['$\Gamma$', 'X', 'M', '$\Gamma$', 'R'],
            cell=model.atoms.cell,
            supercell_matrix=None,
            npoints=100,
            efermi=None,
            erange=None,
            fullband_color="blue",
            downfolded_band_color="green",
            marker="o",
            ax=None,
            savefig="downfold_band.png",
            show=True,
        )
        self.JR_downfolded = ewf.HwannR
        self.Rlist = ewf.Rlist

    def get_JR(self):
        return self.JR_downfolded, self.Rlist


class JDownfolder_pickle:
    def __init__(
        self,
        inpath,
        metals,
        ligands,
        outpath,
        qmesh=[7, 7, 7],
        iso_only=False,
        method="lowdin",
        **kwargs,
    ):
        self.exc = SpinIO.load_pickle(path=inpath, fname="TB2J.pickle")

        self.iso_only = (self.exc.dmi_ddict is None) or iso_only
        self.metals = metals
        self.ligands = ligands
        self.outpath = outpath
        self.method = method
        print("Using method:", self.method)

        # read atomic structure
        self.atoms = self.exc.atoms
        self.nspin = self.exc.nspin
        self.qmesh = qmesh
        self.natom = len(self.atoms)
        self.Rcut = None
        self._build_atom_index()
        self._prepare_distance()
        Jd, Rlist = self._downfold(**kwargs)
        self._Jd_to_exchange(Jd, Rlist)

    def _build_atom_index(self):
        self.magnetic_elements = self.metals
        self.iM = []
        self.iL = []
        self.ind_mag_atoms = []
        for i, sym in enumerate(self.atoms.get_chemical_symbols()):
            if sym in self.magnetic_elements:
                self.iM.append(self.exc.index_spin[i])
                self.ind_mag_atoms.append(i)
            elif sym in self.ligands:
                self.iL.append(self.exc.index_spin[i])

        self.nM = len(self.iM)
        self.nL = len(self.iL)
        self.nsite = self.nM + self.nL

    def _downfold(self, **kwargs):
        JR2 = self.exc.get_full_Jtensor_for_Rlist(asr=True)
        if self.method.lower() == "lowdin":
            d = JDownfolder(
                JR2,
                self.exc.Rlist,
                iM=self.iM,
                iL=self.iL,
                qmesh=self.qmesh,
                iso_only=self.iso_only,
            )
            Jd, Rlist = d.get_JR()
        else:
            d = PWFDownfolder(
                JR2,
                self.exc.Rlist,
                iM=self.iM,
                iL=self.iL,
                qmesh=self.qmesh,
                atoms=self.atoms,
                iso_only=self.iso_only,
                **kwargs,
            )
            Jd, Rlist = d.get_JR()
        return Jd, Rlist

    def _Jd_to_exchange(self, Jd, Rlist):
        self._prepare_distance()
        self._prepare_index_spin()
        self.Jdict = {}
        if self.iso_only:
            self.DMIdict = None
            self.Janidict = None
        else:
            self.DMIdict = {}
            self.Janidict = {}

        for iR, R in enumerate(Rlist):
            for i, ispin in enumerate(self.index_spin):
                for j, jspin in enumerate(self.index_spin):
                    if ispin >= 0 and jspin >= 0:
                        if not (tuple(R) == (0, 0, 0) and ispin == jspin):
                            # self interaction.
                            J33 = Jd[
                                iR, ispin * 3 : ispin * 3 + 3, jspin * 3 : jspin * 3 + 3
                            ]
                            J, DMI, Jani = decompose_J_tensor(J33)
                            self.Jdict[(tuple(R), ispin, jspin)] = J
                            if not self.iso_only:
                                self.DMIdict[(tuple(R), ispin, jspin)] = DMI
                                self.Janidict[(tuple(R), ispin, jspin)] = Jani

        io = SpinIO(
            atoms=self.atoms,
            spinat=self.exc.spinat,
            charges=self.exc.charges,
            index_spin=self.index_spin,
            colinear=self.iso_only,
            distance_dict=self.distance_dict,
            exchange_Jdict=self.Jdict,
            dmi_ddict=self.DMIdict,
            Jani_dict=self.Janidict,
        )

        io.write_all(self.outpath)

    def _prepare_distance(self):
        self.distance_dict = {}
        self.short_Rlist = []
        self.R_ijatom_dict = defaultdict(lambda: [])
        ind_matoms = self.ind_mag_atoms
        for ispin, iatom in enumerate(ind_matoms):
            for jspin, jatom in enumerate(ind_matoms):
                for R in self.exc.Rlist:
                    pos_i = self.atoms.get_positions()[iatom]
                    pos_jR = self.atoms.get_positions()[jatom] + np.dot(
                        R, self.atoms.get_cell()
                    )
                    vec = pos_jR - pos_i
                    distance = np.sqrt(np.sum(vec**2))
                    if self.Rcut is None or distance < self.Rcut:
                        self.distance_dict[(tuple(R), ispin, jspin)] = (vec, distance)
                        self.R_ijatom_dict[tuple(R)].append((iatom, jatom))
        self.short_Rlist = list(self.R_ijatom_dict.keys())

    def _prepare_index_spin(self):
        # index_spin: index in spin hamiltonian of atom. starts from 1. -1 means not considered.
        ind_matoms = []
        self.index_spin = []
        ispin = 0
        for i, sym in enumerate(self.atoms.get_chemical_symbols()):
            if sym in self.magnetic_elements:
                ind_matoms.append(i)
                self.index_spin.append(ispin)
                ispin += 1
            else:
                self.index_spin.append(-1)


def test():
    # pass
    # inpath = "/home/hexu/projects/NiCl2/vasp_inputs/TB2J_results"
    inpath = "/home/hexu/projects/TB2J_example/CrI3/TB2J_results"
    # inpath = "/home/hexu/projects/TB2J_projects/NiCl2/TB2J_NiCl/TB2J_results"
    _fname = os.path.join(inpath, "TB2J.pickle")
    _p = JDownfolder_pickle(
        inpath=inpath, metals=["Ni"], ligands=["Cl"], outpath="TB2J_results_downfolded"
    )


if __name__ == "__main__":
    test()



#!/usr/bin/env python3
from TB2J.magnon import plot_tb2j_magnon_bands


def plot_tb2j_magnon_bands_cli():
    """Command line interface for plotting magnon band structures from TB2J data."""
    import argparse

    parser = argparse.ArgumentParser(
        description="Plot magnon band structure from TB2J data",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "-f", "--file", default="TB2J.pickle", help="Path to TB2J pickle file"
    )

    parser.add_argument(
        "-p",
        "--path",
        help="High-symmetry k-point path (e.g., 'GXMG' for square lattice)",
    )

    parser.add_argument(
        "-n",
        "--npoints",
        type=int,
        default=300,
        help="Number of k-points for band structure calculation",
    )

    parser.add_argument(
        "-a",
        "--anisotropic",
        action="store_true",
        help="Include anisotropic interactions",
    )

    parser.add_argument(
        "-q",
        "--quadratic",
        action="store_true",
        help="Include biquadratic interactions",
    )

    parser.add_argument("-o", "--output", help="Output filename for the plot")

    parser.add_argument(
        "--no-pbc",
        nargs="+",
        type=int,
        choices=[0, 1, 2],
        help="Disable periodic boundary conditions in specified directions (0=x, 1=y, 2=z)",
    )

    args = parser.parse_args()

    # Set up periodic boundary conditions
    pbc = [True, True, True]
    if args.no_pbc:
        for direction in args.no_pbc:
            pbc[direction] = False
    pbc = tuple(pbc)

    # Plot the bands
    plot_tb2j_magnon_bands(
        pickle_file=args.file,
        path=args.path,
        npoints=args.npoints,
        anisotropic=args.anisotropic,
        quadratic=args.quadratic,
        pbc=pbc,
        filename=args.output,
    )


if __name__ == "__main__":
    plot_tb2j_magnon_bands_cli()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
class for basis set
"""
from typing import Any
import numpy as np
import dataclasses
from TB2J.utils import symbol_number


@dataclasses.dataclass
class Basis:
    iatom: int = 0
    symbol: str = "unamed"
    spin: int = 0

    def __str__(self):
        return f"{self.iatom}|{self.symbol}|{self.spin}"

    def to_symnum_type(self, atoms=None, symbol_number=None):
        if symbol_number is None:
            symbol_number = symbol_number(atoms)
        return f"{self.iatom}|{self.symbol}|{self.spin}"


@dataclasses.dataclass
class NAOBasis(Basis):
    n: int = 0
    l: int = 0
    m: int = 0
    zeta: int = 0
    element: str = "unknown"

    def __str__(self) -> str:
        return super().__str__()


# @dataclasses.dataclass
#  inherit from tuple, the elements are basis.


class BasisSet(list):
    def set_atoms(self, atoms):
        self._atoms = atoms

    def get_iorbs_of_atom(self, iatom: int):
        """
        get the index of orbitals of an atom
        """
        return [i for i, basis in enumerate(self) if basis.iatom == iatom]

    def get_iorbs_of_atom_spin(self, iatom: int, spin: int):
        """
        get the index of orbitals of an atom
        """
        return [
            i
            for i, basis in enumerate(self)
            if basis.iatom == iatom and basis.spin == spin
        ]

    def find_iorbs(self, key=None):
        """
        find the index of orbitals of an atom
        """
        return [i for i, basis in enumerate(self) if key(basis)]



import os

import numpy as np

from TB2J.exchange import ExchangeNCL
from TB2J.gpaw_wrapper import GPAWWrapper
from TB2J.utils import auto_assign_basis_name


def gen_exchange_gpaw(
    gpw_fname,
    magnetic_elements=[],
    kmesh=[3, 3, 3],
    emin=-12.0,
    emax=0.0,
    nz=50,
    exclude_orbs=[],
    Rcut=None,
    use_cache=False,
    output_path="TB2J_results",
    description="",
):
    print("Reading from GPAW data and calculate electronic structure.")
    model = GPAWWrapper(gpw_fname=gpw_fname)
    efermi = model.calc.get_fermi_level()
    print(f"Fermi Energy: {efermi}")
    poses = np.vstack([model.positions, model.positions])
    basis, _ = auto_assign_basis_name(
        poses,
        model.atoms,
        write_basis_file=os.path.join(output_path, "assigned_basis.txt"),
    )

    if model.calc.get_spin_polarized():
        print("Starting to calculate exchange.")
        exchange = ExchangeNCL(
            tbmodels=model,
            atoms=model.atoms,
            efermi=efermi,
            basis=basis,
            magnetic_elements=magnetic_elements,
            kmesh=kmesh,
            emin=emin,
            emax=emax,
            nz=nz,
            exclude_orbs=exclude_orbs,
            Rcut=Rcut,
            use_cache=use_cache,
            output_path=output_path,
            description=description,
        )
        exchange.run(path=output_path)
        print("\n")
        print(f"All calculation finished. The results are in {output_path} directory.")



"""
utilities to handel orbital magnetic moments
"""

import numpy as np


def complex_spherical_harmonic_to_real_spherical_harmonic(l=1):
    """
    matrix to convert the complex spherical harmonics to real spherical harmonics
    """
    MCR = np.zeros((2 * l + 1, 2 * l + 1), dtype=np.complex128)
    MCR[0 + l, 0 + l] = 1.0
    for m in range(1, l + 1):
        mi = m + l
        mpi = -m + l
        MCR[mi, mi] = (-1) ** m / 2
        MCR[mi, mpi] = 1.0 / 2
        MCR[mpi, mi] = -1j * (-1) ** m / 2
        MCR[mpi, mpi] = 1j / 2

    return MCR


def test_complex_spherical_harmonic_to_real_spherical_harmonic():
    """
    test the conversion matrix
    """
    l = 3
    MCR = complex_spherical_harmonic_to_real_spherical_harmonic(l)
    # print(MCR*np.sqrt(2))
    print(MCR * 2)


if __name__ == "__main__":
    test_complex_spherical_harmonic_to_real_spherical_harmonic()



"""
This file contains the parser of the density matrix generated by TB2J.
"""

import pickle
import numpy as np
from dataclasses import dataclass


@dataclass
class DensityMatrix:
    """
    Rlist: list of R vectors, np.ndarray of shape (NR, 3), or None
    rhoR: density matrix, np.ndarray of shape (nR, N, N), or None
    Rmap: dict of R vectors, or None
    """

    Rlist: np.ndarray
    rhoR: np.ndarray
    Rmap: dict

    def __init__(self, Rlist=None, rhoR=None):
        self.Rlist = Rlist
        self.rhoR = rhoR
        self._build_Rmap()

    def _build_Rmap(self):
        self.Rmap = {}
        for R in self.Rlist:
            self.Rmap[tuple(R)] = len(self.Rmap)

    def print_dm_at_R(self, R):
        """
        print the real part with 5 decimals of the density matrix at R
        """
        print(np.real(self.rhoR[self.Rmap[tuple(R)]]).round(5))

    def print_dm_at_R_block(self, R, ilist, jlist):
        """
        print the real part with 3 decimals of the density matrix at R, only the block of ilist and jlist
        """
        print("real part")
        print(np.real(self.rhoR[self.Rmap[tuple(R)]][np.ix_(ilist, jlist)]).round(5))
        print("imag part")
        print(np.imag(self.rhoR[self.Rmap[tuple(R)]][np.ix_(ilist, jlist)]).round(5))

    def get_dm_at_R(self, R):
        return self.rhoR[self.Rmap[tuple(R)]]

    @classmethod
    def load_pickle(self, fname):
        with open(fname, "rb") as f:
            d = pickle.load(f)
        return DensityMatrix(**d)



import sisl


def rotate_siesta_DM(DM, noncollinear=False):
    angles_list = [[0.0, 90.0, 0.0], [0.0, 90.0, 90.0]]
    if noncollinear:
        angles_list += [[0.0, 45.0, 0.0], [0.0, 90.0, 45.0], [0.0, 45.0, 90.0]]

    for angles in angles_list:
        yield DM.spin_rotate(angles)


def read_label(fdf_fname):
    label = "siesta"
    with open(fdf_fname, "r") as File:
        for line in File:
            corrected_line = line.lower().replace(".", "").replace("-", "")
            if "systemlabel" in corrected_line:
                label = line.split()[1]
                break

    return label


def rotate_DM(fdf_fname, noncollinear=False):
    fdf = sisl.get_sile(fdf_fname)
    DM = fdf.read_density_matrix()
    label = read_label(fdf_fname)

    rotated = rotate_siesta_DM(DM, noncollinear=noncollinear)

    for i, rotated_DM in enumerate(rotated):
        rotated_DM.write(f"{label}_{i+1}.DM")
    DM.write(f"{label}_0.DM")

    print(
        f"The output has been written to the {label}_i.DM files. {label}_0.DM contains the reference density matrix."
    )



"""
This module calculate Oiju from electron-phonon coupling matrix and Electron Wannier function.
"""

import numpy as np
import progressbar
from .exchangeCL2 import ExchangeCL2
from collections import defaultdict


class OijuWannEPC(ExchangeCL2):
    def __init__(
        self,
        tbmodels,
        atoms,
        efermi,
        basis=None,
        magnetic_elements=[],
        kmesh=[4, 4, 4],
        emin=-15,  # integration lower bound, relative to fermi energy
        emax=0.05,  # integration upper bound. Should be 0 (fermi energy). But DFT codes define Fermi energy in various ways.
        exclude_orbs=[],  #
        Rmesh=[0, 0, 0],  # Rmesh.
        description="",
        qmesh=[2, 2, 2],
        EPCmats=None,
        WannUmats=None,
    ):
        super(self).__init__(
            tbmodels,
            atoms,
            efermi,
            basis=None,
            magnetic_elements=[],
            kmesh=[4, 4, 4],
            emin=-15,  # integration lower bound, relative to fermi energy
            emax=0.05,  # integration upper bound. Should be 0 (fermi energy).
            # But DFT codes define Fermi energy in various ways.
            exclude_orbs=[],  #
            Rmesh=[0, 0, 0],  # Rmesh.
            description="",
        )

        # prepare EPC in electron wannier space
        self.EPCmat_up, self.EPCmat_dn = EPCmats
        self.Umat_up, self.Umat_dn = WannUmats
        self.prepare_epc_wann()

        # prepare dDelta
        self.calc_dDelta()

        self._kmap = None

    def prepare_epc_wann(self):
        """
        prepare EPC in electron wannier function representation
        """
        self.EPCmat_wann_up = self.EPCmat_up.to_wann(self.Umat_up)
        self.EPCmat_wann_dn = self.EPCmat_dn.to_wann(self.Umat_dn)

    def get_iq(self, q):
        if self._kmap is None:
            self._kmap = {}
        for i, q in enumerate(self.klist):
            self._kmap[tuple(q)] = i

    def calc_dDelta(self):
        """
        calculate $\delta \Delta$.
        <m 0| dV/d \tau(qV) | n 0>. Since only 0->0, q=0)

        \delta\Delta (q, v) = \sum_k w(k) e^{ik R=0} U\dagger (k+q) g(q,v) U(k)
                     =\sum_k w(k) U\dagger g(0, v) U(k)

        For q/=0. \delta\Delta=0.
        the result will be a matrix[nphon, nwann, nwann]
        """
        self.dDelta = np.zeros((self.nphon, self.nwann, self.nwann))
        # iq, iv, iR
        # iq0 = self.iqlist[(0, 0, 0)]
        iq0 = self.get_iq((0, 0, 0))

        for iv in range(self.nphon):
            for ik, k in enumerate(self.klist):
                self.dDelta[iv] += self.kweight[ik] * (
                    self.EPCmat_wann_up[iq0, iv, ik, :, :]
                    - self.EPCmat_wann_dn[iq0, iv, ik, :, :]
                )

    def deltaG_kvq(self, Gk, Umat, EPCmat):
        pass

    def get_Oijvq(self, GR_up, GR_dn, dGR_vq_up, dGR_vq_dn, iatom, jatom, Rj):
        """
        calculate Oiju(i, j, Rj, v, q)
        """
        iorbs = self.iorb(iatom)
        jorbs = self.iorb(jatom)

        self.get_Delta(iatom) @ dGR_vq_up[np.ix_(iorbs, jorbs)] @ self.get_Delta(
            jatom
        ) @ GR_dn

    def _build_Rjdict(self, Rjlist):
        for iRj, Rj in enumerate(Rjlist):
            self.Rudict[tuple(Rj)] = iRj

    def _build_Rudict(self, Rulist):
        for iRu, Ru in enumerate(Rulist):
            self.Rudict[tuple(Ru)] = iRu

    def iRu(self, Ru):
        return self.Rudict[tuple(Ru)]

    def iRj(self, Rj):
        return self.Rjdict[tuple(Rj)]

    def Oij_vq(self, Rj_list, energy):
        GR_up = defaultdict(lambda: 0.0)
        GR_dn = defaultdict(lambda: 0.0)
        dGR_vq_up = defaultdict(lambda: 0.0)
        dGR_vq_dn = defaultdict(lambda: 0.0)
        self.Oijvq = defaultdict(lambda: 0.0)

        for iq, q in enumerate(self.qmesh):
            for v in self.vlist:
                # calculate dG(i,j, Rj, v, q)
                for ik, k in enumerate(self.klist):
                    # Gk
                    Gk_up = self.Gup.get_Gk(ik, energy)
                    Gk_dn = self.Gdn.get_Gk(ik, energy)

                    # GR
                    for iR, Rj in enumerate(Rj_list):
                        Rj = np.array(Rj)
                        phase = np.exp(-2j * np.pi * np.dot(Rj, k))
                        GR_up[self.iRj(Rj)] += Gk_up * (phase * self.kweights[ik])
                        GR_dn[self.iRj(Rj)] += Gk_dn * (phase * self.kweights[ik])

                    # deltaG(k, v, q, energy)
                    dGk_vq_up = self.deltaG_kvq(Gk_up, self.EPCmat_wann_up)
                    dGk_vq_dn = self.deltaG_kvq(Gk_dn, self.EPCmat_wann_dn)

                    for iR, Rj in enumerate(Rj_list):
                        phase = np.exp(-2j * np.pi * np.dot(Rj, k))
                        # nwann*nwann matrix
                        dGR_vq_up[Rj] += dGk_vq_up * (phase * self.kweights[ik])
                        dGR_vq_dn[Rj] += dGk_vq_dn * (phase * self.kweights[ik])

                    for iatom in self.iatom_list:
                        for jatom in self.iatom_list:
                            self.Oijvq[(iatom, jatom, Rj, v, q)] += self.get_Oijvq(
                                GR_up,
                                GR_dn,
                                dGR_vq_up,
                                dGR_vq_dn,
                                iatom,
                                jatom,
                                Rj,
                            )
                            for iRu, Ru in enumerate(self.Rulist):
                                phase_Ru = np.exp(-2j * np.pi * np.dot(q, Ru))
                                # u=v for atomic displacement.
                                self.Oiju[(iatom, jatom, Rj, v, Ru)] += (
                                    phase_Ru * self.Oijvq[(iatom, jatom, Rj, v, q)]
                                )

    def calculate_all(self):
        print("Green's function Calculation started.")
        widgets = [
            " [",
            progressbar.Timer(),
            "] ",
            progressbar.Bar(),
            " (",
            progressbar.ETA(),
            ") ",
        ]
        bar = progressbar.ProgressBar(maxval=len(self.elist) - 1, widgets=widgets)
        bar.start()
        for ie in range(len(self.elist[:-1])):
            bar.update(ie)
            em, en = self.elist[ie], self.elist[ie + 1]
            e = (em + en) / 2
            de = en - em
            # Green's function for G(i, j, Rj)
            GR_up = self.Gup.get_GR(self.Rlist, energy=e)
            GR_dn = self.Gdn.get_GR(self.Rlist, energy=e)
            self.get_rho_e(GR_up, GR_dn, de)
            self.get_all_A(GR_up, GR_dn, de)
        self.get_rho_atom()
        self.A_to_Jtensor()
        bar.finish()

    def run(self):
        pass



# -*- coding: utf-8 -*-
# File: mycfr.py
# Time: 2017/12/10 19:29:43
"""
Continued fraction representation.
"""

import numpy as np
from scipy.linalg import eig

kb = 8.61733e-5  # Boltzmann constant in eV


class CFR:
    """
    Integration with the continued fraction representation.
    """

    def __init__(self, nz=50, T=60):
        self.nz = nz
        self.T = 600
        self.beta = 1 / (kb * self.T)
        self.Rinf = 1e10
        if nz <= 0:
            raise ValueError("nz should be larger than 0.")
        else:
            self.prepare_poles()

    def prepare_poles(self):
        ##b_mat = [1 / (2.0 * np.sqrt((2 * (j + 1) - 1) * (2 * (j + 1) + 1)) / (kb * self.#T)) for j in range(0, self.nz- 1)]
        jmat = np.arange(0, self.nz - 1)
        b_mat = 1 / (2.0 * np.sqrt((2 * (jmat + 1) - 1) * (2 * (jmat + 1) + 1)))
        b_mat = np.diag(b_mat, -1) + np.diag(b_mat, 1)
        self.poles, residules = eig(b_mat)

        residules = 0.25 * np.abs(residules[0, :]) ** 2 / self.poles**2
        # residules = 0.25 * np.diag(residules) ** 2 / self.poles**2

        # self.weights = np.where(
        #        #np.real(self.poles) > 0, 4.0j / self.beta * residules, 0.0
        #        #np.real(self.poles) > 0, 2.0j / self.beta * residules, 2.0j / self.beta * residules
        #    np.real(self.poles) > 0, 2.0j / self.beta * residules, 0.0
        # )

        # self.path = 1j / self.poles * kb * self.T

        self.path = []
        self.weights = []
        for p, r in zip(self.poles, residules):
            if p > 0:
                self.path.append(1j / p * kb * self.T)
                w = 2.0j / self.beta * r
                self.weights.append(w)
                self.path.append(-1j / p * kb * self.T)
                self.weights.append(w)

        self.path = np.array(self.path)
        self.weights = np.array(self.weights)

        # from J. Phys. Soc. Jpn. 88, 114706 (2019)
        # A_mat =  -1/2 *np.diag(1, -1) + np.diag(1, 1)
        # B_mat = np.diag([2*i-1 for i in range(1, self.nz)])
        # eigp, eigv = eig(A_mat, B_mat)
        # zp = 1j / eigp * kb * self.T
        # Rp = 0.25 * np.diag(eigv)**2 * zp **2

        # print the poles and the weights
        for i in range(len(self.poles)):
            print("Pole: ", self.poles[i], "Weight: ", self.weights[i])

        # add a point to the poles: 1e10j
        self.path = np.concatenate((self.path, [self.Rinf * 1j]))
        # self.path = np.concatenate((self.path, [0.0]))
        self.npoles = len(self.poles)

        # add weights for the point 1e10j
        # self.weights = np.concatenate((self.weights, [-self.Rinf]))
        self.weights = np.concatenate((self.weights, [00.0]))
        # zeros moment is 1j * R * test_gf(1j * R), but the real part of it will be taken. In contrast to the other part, where the imaginary part is taken.

    def integrate_values(self, gf_vals, imag=False):
        ret = 0
        for i in range(self.npoles + 1):
            ret += self.weights[i] * gf_vals[i]
        ret *= (
            -np.pi / 2
        )  # This is to be compatible with the integration of contour, where /-np.pi is used after the integration. And the factor of 2 is because the Fermi function here is 2-fold degenerate.
        if imag:
            return np.imag(ret)
        return ret

    def integrate_func(self, gf, ef=0):
        """
        Integration with the continued fraction representation.

        :param gf: Green's function
        :param ef: Fermi energy
        :return: integration result
        """
        path = self.path
        gf_vals = gf(path, ef=ef)
        return self.integrate_values(gf_vals)


def test_cfr():
    cfr = CFR(nz=100)

    def test_gf(z, ef=0.1):
        return 1 / (z - 3 + ef)

    r = cfr.integrate_func(test_gf, ef=2)
    r = -np.imag(r) / np.pi * 2
    print(r)
    return r


if __name__ == "__main__":
    test_cfr()



# -*- coding: utf-8 -*-
import re
import struct

import numpy as np
from scipy.sparse import csr_matrix


class XR_matrix:
    def __init__(self, nspin, XR_fileName):
        self.nspin = nspin
        self.XR_fileName = XR_fileName

    def read_file(self):
        with open(self.XR_fileName, "r") as fread:
            while True:
                line = fread.readline().split()
                if line[0] == "Matrix":
                    break

            self.basis_num = int(line[-1])
            line = fread.readline()
            self.R_num = int(line.split()[-1])
            self.R_direct_coor = np.zeros([self.R_num, 3], dtype=int)
            if self.nspin != 4:
                self.XR = np.zeros(
                    [self.R_num, self.basis_num, self.basis_num], dtype=float
                )
            else:
                self.XR = np.zeros(
                    [self.R_num, self.basis_num, self.basis_num], dtype=complex
                )

            for iR in range(self.R_num):
                line = fread.readline().split()
                self.R_direct_coor[iR, 0] = int(line[0])
                self.R_direct_coor[iR, 1] = int(line[1])
                self.R_direct_coor[iR, 2] = int(line[2])
                data_size = int(line[3])

                if self.nspin != 4:
                    data = np.zeros((data_size,), dtype=float)
                else:
                    data = np.zeros((data_size,), dtype=complex)

                indices = np.zeros((data_size,), dtype=int)
                indptr = np.zeros((self.basis_num + 1,), dtype=int)

                if data_size != 0:
                    if self.nspin != 4:
                        line = fread.readline().split()
                        if len(line) != data_size:
                            print("size = ", len(line), " data_size = ", data_size)
                        for index in range(data_size):
                            data[index] = float(line[index])
                    else:
                        line = re.findall("[(](.*?)[])]", fread.readline())
                        for index in range(data_size):
                            value = line[index].split(",")
                            data[index] = complex(float(value[0]), float(value[1]))

                    line = fread.readline().split()
                    for index in range(data_size):
                        indices[index] = int(line[index])

                    line = fread.readline().split()
                    for index in range(self.basis_num + 1):
                        indptr[index] = int(line[index])

                self.XR[iR] = csr_matrix(
                    (data, indices, indptr), shape=(self.basis_num, self.basis_num)
                ).toarray()

    def read_file_binary(self):
        with open(self.XR_fileName, "rb") as fread:
            self.basis_num = struct.unpack("i", fread.read(4))[0]
            self.R_num = struct.unpack("i", fread.read(4))[0]
            self.R_direct_coor = np.zeros([self.R_num, 3], dtype=int)
            if self.nspin != 4:
                self.XR = np.zeros(
                    [self.R_num, self.basis_num, self.basis_num], dtype=float
                )
            else:
                self.XR = np.zeros(
                    [self.R_num, self.basis_num, self.basis_num], dtype=complex
                )

            for iR in range(self.R_num):
                self.R_direct_coor[iR, 0] = struct.unpack("i", fread.read(4))[0]
                self.R_direct_coor[iR, 1] = struct.unpack("i", fread.read(4))[0]
                self.R_direct_coor[iR, 2] = struct.unpack("i", fread.read(4))[0]
                data_size = struct.unpack("i", fread.read(4))[0]

                if self.nspin != 4:
                    data = np.zeros((data_size,), dtype=float)
                else:
                    data = np.zeros((data_size,), dtype=complex)

                indices = np.zeros((data_size,), dtype=int)
                indptr = np.zeros((self.basis_num + 1,), dtype=int)

                if data_size != 0:
                    if self.nspin != 4:
                        for index in range(data_size):
                            data[index] = struct.unpack("d", fread.read(8))[0]
                    else:
                        for index in range(data_size):
                            real = struct.unpack("d", fread.read(8))[0]
                            imag = struct.unpack("d", fread.read(8))[0]
                            data[index] = complex(real, imag)

                    for index in range(data_size):
                        indices[index] = struct.unpack("i", fread.read(4))[0]

                    for index in range(self.basis_num + 1):
                        indptr[index] = struct.unpack("i", fread.read(4))[0]

                self.XR[iR] = csr_matrix(
                    (data, indices, indptr), shape=(self.basis_num, self.basis_num)
                ).toarray()


def read_HR_SR(
    nspin=4,
    binary=False,
    HR_fileName="data-HR-sparse_SPIN0.csr",
    SR_fileName="data-SR-sparse_SPIN0.csr",
):
    """
    IN:
        nspin: int, different spins.
        binary: bool, whether the HR and SR matrices are binary files.
        HR_fileName: if nspin=1 or 4, str, HR file name;
                     if nspin=2, list or tuple, size=2, [HR_up_fileName, HR_dn_fileName].
        SR_fileName: str, SR file name.

    OUT:
        if nspin = 1 or 4:
            return basis_num, R_direct_coor, HR, SR
        elif nspin = 2:
            return basis_num, R_direct_coor, HR_up, HR_dn, SR

        basis_num: int, number of atomic orbital basis.
        R_direct_coor: numpy ndarray, shape=[R_num, 3], fractional coordinates (x, y, z) of the R-th primitive cell.
        HR or HR_up or HR_dn: numpy ndarray, shape=[R_num, basis_num, basis_num], HR matrix and unit is eV.
        SR: numpy ndarray, shape=[R_num, basis_num, basis_num], SR matrix.
    """
    Ry_to_eV = 13.605698066

    if nspin == 1 or nspin == 4:
        if not isinstance(HR_fileName, str):
            raise ValueError("The HR_fileName must be a str for nspin=1 or 4.")

        HR = XR_matrix(nspin, HR_fileName)
        SR = XR_matrix(nspin, SR_fileName)

        if binary:
            HR.read_file_binary()
            SR.read_file_binary()
        else:
            HR.read_file()
            SR.read_file()

        return HR.basis_num, HR.R_direct_coor, HR.XR * Ry_to_eV, SR.XR
    else:
        if not isinstance(HR_fileName, (list, tuple)):
            raise ValueError("The HR_fileName must be a list or a tuple for nspin=2.")

        if len(HR_fileName) != 2:
            raise ValueError("The size of the HR_fileName must be 2 for nspin=2.")

        HR_up = XR_matrix(nspin, HR_fileName[0])
        HR_dn = XR_matrix(nspin, HR_fileName[1])
        SR = XR_matrix(nspin, SR_fileName)

        if binary:
            HR_up.read_file_binary()
            HR_dn.read_file_binary()
            SR.read_file_binary()
        else:
            HR_up.read_file()
            HR_dn.read_file()
            SR.read_file()

        return (
            HR_up.basis_num,
            HR_up.R_direct_coor,
            HR_up.XR * Ry_to_eV,
            HR_dn.XR * Ry_to_eV,
            SR.XR,
        )



import argparse
from dataclasses import dataclass

import yaml

__all__ = ["ExchangeParams", "add_exchange_args_to_parser", "parser_argument_to_dict"]


@dataclass
class ExchangeParams:
    """
    A class to store the parameters for exchange calculation.
    """

    efermi: float
    basis = []
    magnetic_elements = []
    index_magnetic_atoms = None
    include_orbs = {}
    _kmesh = [4, 4, 4]
    emin: float = -15
    emax: float = 0.05
    nz: int = 100
    exclude_orbs = []
    ne: int = 0
    Rcut: float = None
    _use_cache: bool = False
    nproc: int = 1
    description: str = ""
    write_density_matrix: bool = False
    orb_decomposition: bool = False
    output_path: str = "TB2J_results"
    mae_angles = None
    orth = False
    ibz = False

    def __init__(
        self,
        efermi=-10.0,
        basis=None,
        magnetic_elements=None,
        include_orbs=None,
        kmesh=[4, 4, 4],
        emin=-15,
        emax=0.00,
        nz=100,
        ne=None,
        Rcut=None,
        use_cache=False,
        nproc=1,
        description="",
        write_density_matrix=False,
        orb_decomposition=False,
        output_path="TB2J_results",
        exclude_orbs=[],
        mae_angles=None,
        orth=False,
        ibz=False,
        index_magnetic_atoms=None,
    ):
        self.efermi = efermi
        self.basis = basis
        # self.magnetic_elements = magnetic_elements
        # self.include_orbs = include_orbs
        self.magnetic_elements, self.include_orbs = self.set_magnetic_elements(
            magnetic_elements, include_orbs
        )
        self._kmesh = kmesh
        self.emin = emin
        self.emax = emax
        self.nz = nz
        self.exclude_orbs = exclude_orbs
        self.ne = ne
        self.Rcut = Rcut
        self._use_cache = use_cache
        self.nproc = nproc
        self.description = description
        self.write_density_matrix = write_density_matrix
        self.orb_decomposition = orb_decomposition
        self.output_path = output_path
        self.mae_angles = mae_angles
        self.orth = orth
        self.ibz = ibz
        self.index_magnetic_atoms = index_magnetic_atoms

    def set_params(self, **kwargs):
        for key, val in kwargs.items():
            setattr(self, key, val)

    def save_to_yaml(self, fname):
        with open(fname, "w") as myfile:
            yaml.dump(self.__dict__, myfile)

    def set_magnetic_elements(self, magnetic_elements, include_orbs):
        # magnetic_elements = exargs.pop("magnetic_elements")
        # include_orbs = exargs.pop("include_orbs")
        if include_orbs is None:
            include_orbs = {}
        if isinstance(magnetic_elements, str):
            magnetic_elements = [magnetic_elements]
        for element in magnetic_elements:
            if "_" in element:
                elem = element.split("_")[0]
                orb = element.split("_")[1:]
                include_orbs[elem] = orb
            else:
                include_orbs[element] = None

        magnetic_elements = list(include_orbs.keys())
        return magnetic_elements, include_orbs


def add_exchange_args_to_parser(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--elements",
        help="elements to be considered in Heisenberg model",
        default=None,
        type=str,
        nargs="*",
    )

    parser.add_argument(
        "--spinor",
        help="whether the Wannier functions are spinor. Default: False",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "--rcut",
        help="cutoff of spin pair distance. The default is to calculate all commensurate R point to the k mesh.",
        default=None,
        type=float,
    )
    parser.add_argument("--efermi", help="Fermi energy in eV", default=None, type=float)
    parser.add_argument(
        "--ne",
        help="number of electrons in the unit cell. If not given, TB2J will use the fermi energy to compute it.",
    )
    parser.add_argument(
        "--kmesh",
        help="kmesh in the format of kx ky kz",
        type=int,
        nargs="*",
        default=[5, 5, 5],
    )
    parser.add_argument(
        "--emin",
        help="energy minimum below efermi, default -14 eV",
        type=float,
        default=-14.0,
    )
    parser.add_argument(
        "--emax",
        help="energy maximum above efermi, default 0.0 eV",
        type=float,
        default=0.0,
    )
    parser.add_argument(
        "--nz",
        help="number of steps for semicircle contour, default: 100",
        default=100,
        type=int,
    )
    parser.add_argument(
        "--cutoff",
        help="The minimum of J amplitude to write, (in eV), default is 1e-5 eV",
        default=1e-5,
        type=float,
    )
    parser.add_argument(
        "--exclude_orbs",
        help="the indices of wannier functions to be excluded from magnetic site. counting start from 0",
        default=[],
        type=int,
        nargs="+",
    )

    parser.add_argument(
        "-np",
        help="number of cpu cores to use in parallel, default: 1",
        default=1,
        type=int,
    )

    parser.add_argument(
        "--use_cache",
        help="whether to use disk file for temporary storing wavefunctions and hamiltonian to reduce memory usage. Default: False",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "--description",
        help="add description of the calculatiion to the xml file. Essential information, like the xc functional, U values, magnetic state should be given.",
        type=str,
        default="Calculated with TB2J.",
    )

    parser.add_argument(
        "--orb_decomposition",
        default=False,
        action="store_true",
        help="whether to do orbital decomposition in the non-collinear mode.",
    )

    parser.add_argument(
        "--output_path",
        help="The path of the output directory, default is TB2J_results",
        type=str,
        default="TB2J_results",
    )
    parser.add_argument(
        "--write_dm",
        help="whether to write density matrix",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "--orth",
        help="whether to use lowdin orthogonalization before diagonalization (for testing only)",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "--ibz",
        help=" use irreducible k-points in the Brillouin zone. (Note: only for computing total MAE).",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "--mae_angles",
        help="angles for computing MAE, default is 0 0 0",
        type=float,
        nargs="*",
        default=[0.0, 0.0, 0.0],
    )
    parser.add_argument(
        "--index_magnetic_atoms",
        help="index of magnetic atoms in the unit cell, default is None. If specified, this will be used to determine the atoms to be considered as magnetic atoms, instead of determined from magnetic elements. Note that the index starts from 1 ",
        type=int,
        nargs="*",
        default=None,
    )

    return parser


def parser_argument_to_dict(args) -> dict:
    ind_mag_atoms = args.index_magnetic_atoms
    if ind_mag_atoms is not None:
        ind_mag_atoms = [int(i) - 1 for i in ind_mag_atoms]
    else:
        ind_mag_atoms = None
    return {
        "efermi": args.efermi,
        "magnetic_elements": args.elements,
        "kmesh": args.kmesh,
        "emin": args.emin,
        "emax": args.emax,
        "nz": args.nz,
        "exclude_orbs": args.exclude_orbs,
        "ne": args.ne,
        "Rcut": args.rcut,
        "use_cache": args.use_cache,
        "nproc": args.np,
        "description": args.description,
        "write_density_matrix": args.write_dm,
        "orb_decomposition": args.orb_decomposition,
        "output_path": args.output_path,
        "orth": args.orth,
        "index_magnetic_atoms": ind_mag_atoms,
    }



import argparse
import os
import sys

from HamiltonIO.lawaf import LawafHamiltonian as Ham

from TB2J.exchange_params import add_exchange_args_to_parser, parser_argument_to_dict
from TB2J.versioninfo import print_license

from .manager import Manager


class LaWaFManager(Manager):
    def __init__(
        self,
        path,
        prefix_up,
        prefix_dn,
        prefix_SOC,
        colinear=True,
        wannier_type="LaWaF",
        **kwargs,
    ):
        # models and basis
        if colinear:
            tbmodels, basis, atoms = self.prepare_model_colinear(
                path, prefix_up, prefix_dn
            )
        else:
            tbmodels, basis, atoms = self.prepare_model_ncl(path, prefix_SOC)

        description = self.description(path, prefix_up, prefix_dn, prefix_SOC, colinear)
        kwargs["description"] = description

        super().__init__(atoms, tbmodels, basis, colinear=colinear, **kwargs)

    def prepare_model_colinear(self, path, prefix_up, prefix_dn):
        print("Reading LawaF hamiltonian: spin up.")
        tbmodel_up = Ham.load_pickle(os.path.join(path, f"{prefix_up}.pickle"))
        print("Reading LaWaF hamiltonian: spin down.")
        tbmodel_dn = Ham.load_pickle(os.path.join(path, f"{prefix_dn}.pickle"))
        tbmodels = (tbmodel_up, tbmodel_dn)
        basis = tbmodel_up.wann_names
        atoms = tbmodel_up.atoms
        return tbmodels, basis, atoms

    def prepare_model_ncl(self, path, prefix_SOC):
        print("Reading LaWaF hamiltonian: non-colinear spin.")
        tbmodel = Ham.load_pickle(os.path.join(path, f"{prefix_SOC}.pickle"))
        basis = tbmodel.wann_names
        atoms = tbmodel.atoms
        return tbmodel, basis, atoms

    def description(self, path, prefix_up, prefix_dn, prefix_SOC, colinear=True):
        if colinear:
            description = f""" Input from collinear LaWaF data.
 Tight binding data from {path}.
 Prefix of wannier function files:{prefix_up} and {prefix_dn}.
Warning: Please check if the noise level of Wannier function Hamiltonian to make sure it is much smaller than the exchange values.
\n"""

        else:
            description = f""" Input from non-collinear LaWaF data.
Tight binding data from {path}.
Prefix of wannier function files:{prefix_SOC}.
Warning: Please check if the noise level of Wannier function Hamiltonian to make sure it is much smaller than the exchange values.
The DMI component parallel to the spin orientation, the Jani which has the component of that orientation should be disregarded
e.g. if the spins are along z, the xz, yz, zz, zx, zy components and the z component of DMI.
If you need these component, try to do three calculations with spin along x, y, z,  or use structure with z rotated to x, y and z. And then use TB2J_merge.py to get the full set of parameters.\n"""

        return description


def lawaf2J_cli():
    print_license()
    parser = argparse.ArgumentParser(
        description="lawaf2J: Using magnetic force theorem to calculate exchange parameter J from wannier functions in LaWaF"
    )
    parser.add_argument(
        "--path", help="path to the wannier files", default="./", type=str
    )

    parser.add_argument(
        "--spinor",
        help="if the calculation is spinor, default is False",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--prefix_spinor",
        help="prefix to the spinor wannier files",
        default="lawaf_spinor.pickle",
        type=str,
    )
    parser.add_argument(
        "--prefix_up",
        help="prefix to the spin up wannier files",
        default="lawaf_up",
        type=str,
    )
    parser.add_argument(
        "--prefix_down",
        help="prefix to the spin down wannier files",
        default="lawaf_dn",
        type=str,
    )
    add_exchange_args_to_parser(parser)

    args = parser.parse_args()

    if args.efermi is None:
        print("Please input fermi energy using --efermi ")
        sys.exit()
    if args.elements is None:
        print("Please input the magnetic elements, e.g. --elements Fe Ni")
        sys.exit()

    kwargs = parser_argument_to_dict(args)

    manager = LaWaFManager(
        path=args.path,
        prefix_up=args.prefix_up,
        prefix_dn=args.prefix_down,
        prefix_SOC=args.prefix_spinor,
        colinear=not args.spinor,
        **kwargs,
    )
    return manager



#!/usr/bin/env python3
import argparse
import os
import sys
from TB2J.Jdownfolder import JDownfolder_pickle


def main():
    parser = argparse.ArgumentParser(
        description="TB2J_downfold: Downfold the exchange parameter with ligand spin as independent variable to that only with metal site spin as independent variables."
    )
    parser.add_argument(
        "--inpath",
        "-i",
        type=str,
        help="The path of the input TB2J results before the downfolding.",
        default="TB2J_results",
    )

    parser.add_argument(
        "--outpath",
        "-o",
        type=str,
        help="The path of the output TB2J results path after the downfolding",
        default="TB2J_results_downfolded",
    )

    parser.add_argument(
        "--metals",
        "-m",
        type=str,
        help="List of magnetic cation elements",
        nargs="+",
        default=[],
    )

    parser.add_argument(
        "--ligands",
        "-l",
        type=str,
        help="List of ligand elements",
        nargs="+",
        default=[],
    )

    parser.add_argument(
        "--qmesh",
        help="kmesh in the format of kx ky kz",
        type=int,
        nargs="*",
        default=[5, 5, 5],
    )

    parser.add_argument(
        "--iso_only",
        help="whether to downfold only the isotropic part. The other parts will be neglected.",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "--method",
        help="The method to downfold the exchange parameter. Options are Lowdin and PWF (projected Wannier function). ",
        type=str,
        default="Lowdin",
    )

    args = parser.parse_args()

    if len(args.metals) == []:
        print("List of magnetic cation elements cannot be empty")

    if len(args.ligands) == []:
        print("List of ligand elements cannot be empty")

    JDownfolder_pickle(
        inpath=args.inpath,
        metals=args.metals,
        ligands=args.ligands,
        outpath=args.outpath,
        qmesh=args.qmesh,
        iso_only=args.iso_only,
        method=args.method,
    )


main()



from dataclasses import dataclass

import numpy as np
from scipy.spatial.transform import Rotation

from ..io_exchange import SpinIO
from ..mathutils import Hermitize, get_rotation_arrays
from .plot import BandsPlot


@dataclass
class Magnon:
    """
    Magnon calculator implementation using dataclass
    """

    nspin: int
    # ind_atoms: list
    magmom: np.ndarray
    Rlist: np.ndarray
    JR: np.ndarray
    _Q: np.ndarray = np.array([0.0, 0.0, 0.0], dtype=float)
    _uz: np.ndarray = np.array([[0.0, 0.0, 1.0]], dtype=float)
    _n: np.ndarray = np.array([0, 0, 1], dtype=float)

    def set_reference(self, Q, uz, n):
        """
        Set reference propagation vector and quantization axis

        Parameters
        ----------
        Q : array_like
            Propagation vector
        uz : array_like
            Quantization axis
        n : array_like
            Normal vector for rotation
        """
        self.set_propagation_vector(Q)
        self._uz = np.array(uz, dtype=float)
        self._n = np.array(n, dtype=float)

    def set_propagation_vector(self, Q):
        """Set propagation vector"""
        self._Q = np.array(Q)

    @property
    def Q(self):
        """Get propagation vector"""
        if self._Q is None:
            raise ValueError("Propagation vector Q is not set.")
        return self._Q

    @Q.setter
    def Q(self, value):
        if not isinstance(value, (list, np.ndarray)):
            raise TypeError("Propagation vector Q must be a list or numpy array.")
        if len(value) != 3:
            raise ValueError("Propagation vector Q must have three components.")
        self._Q = np.array(value)

    def Jq(self, kpoints):
        """
        Compute the exchange interactions in reciprocal space.

        The exchange interactions J(q) are computed using the Fourier transform:
        J(q) = ∑_R J(R) exp(iq·R)

        Array shapes and indices:
        - kpoints: (nkpt, 3) array of k-points
        - Rlist: (nR, 3) array of real-space lattice vectors
        - JR: (nR, nspin, nspin, 3, 3) array of exchange tensors in real space
            where nspin is number of magnetic atoms
        - Output Jq: (nkpt, nspin, nspin, 3, 3) array of exchange tensors in q-space

        If propagation vector Q is set, each J(R) is rotated before the Fourier transform:
        J'_mn(R) = R_m(ϕ)^T J(R) R_n(ϕ)
        where ϕ = 2π R·Q and R(ϕ) is rotation matrix around axis n by angle ϕ

        Parameters
        ----------
        kpoints : array_like (nkpt, 3)
            k-points at which to evaluate the exchange interactions

        Returns
        -------
        numpy.ndarray (nkpt, nspin, nspin, 3, 3)
            Exchange interaction tensors J(q) at each k-point
            First two indices are for magnetic atom pairs
            Last two indices are for 3x3 tensor components
        """
        Rlist = np.array(self.Rlist)
        JR = self.JR
        JRprime = JR.copy()

        for iR, R in enumerate(Rlist):
            if self._Q is not None:
                # Rotate exchange tensors based on propagation vector
                phi = 2 * np.pi * R @ self._Q  # angle ϕ = 2π R·Q
                rv = phi * self._n  # rotation vector
                Rmat = Rotation.from_rotvec(rv).as_matrix()
                # J'_mn(R) = R_m(ϕ)^T J(R) R_n(ϕ) using Einstein summation.
                # Here m is always in the R=0, thus the rotation is only applied on the
                # n , so only on the right.
                JRprime[iR] = np.einsum(" rijxy, yb -> rijab", JR[iR], Rmat)

        nkpt = kpoints.shape[0]
        Jq = np.zeros((nkpt, self.nspin, self.nspin, 3, 3), dtype=complex)

        for iR, R in enumerate(Rlist):
            for iqpt, qpt in enumerate(kpoints):
                # Fourier transform of exchange tensors
                phase = 2 * np.pi * R @ qpt
                Jq[iqpt] += np.exp(1j * phase) * JRprime[iR]

        # Ensure Hermiticity: J(q) = J(-q)†
        for iqpt in range(nkpt):
            Jq[iqpt, :, :, :, :] += np.conj(
                np.moveaxis(Jq[iqpt, :, :, :, :], [1, 3], [2, 4])
            )
            Jq[iqpt, :, :, :, :] /= 2.0
        return Jq

    def Hq(self, kpoints, anisotropic=True):
        """
        Compute the magnon Hamiltonian in reciprocal space.

        Parameters
        ----------
        kpoints : array_like
            k-points at which to evaluate the Hamiltonian
        anisotropic : bool, optional
            Whether to include anisotropic interactions, default True

        Returns
        -------
        numpy.ndarray
            Magnon Hamiltonian matrix at each k-point
        """
        magmoms = self.magmom.copy()
        magmoms /= np.linalg.norm(magmoms, axis=-1)[:, None]

        U, V = get_rotation_arrays(magmoms, u=self._uz)

        J0 = self.Jq(np.zeros((1, 3)), anisotropic=anisotropic)
        J0 = -Hermitize(J0)[:, :, 0]
        Jq = -Hermitize(self.Jq(kpoints, anisotropic=anisotropic))

        C = np.diag(np.einsum("ix,ijxy,jy->i", V, 2 * J0, V))
        B = np.einsum("ix,ijkxy,jy->kij", U, Jq, U)
        A1 = np.einsum("ix,ijkxy,jy->kij", U, Jq, U.conj())
        A2 = np.einsum("ix,ijkxy,jy->kij", U.conj(), Jq, U)

        return np.block([[A1 - C, B], [B.swapaxes(-1, -2).conj(), A2 - C]])

    def _magnon_energies(self, kpoints, anisotropic=True, u=None):
        """Calculate magnon energies"""
        H = self.Hq(kpoints, anisotropic=anisotropic)
        n = H.shape[-1] // 2
        I = np.eye(n)

        min_eig = 0.0
        try:
            K = np.linalg.cholesky(H)
        except np.linalg.LinAlgError:
            try:
                K = np.linalg.cholesky(H + 1e-6 * np.eye(2 * n))
            except np.linalg.LinAlgError:
                from warnings import warn

                min_eig = np.min(np.linalg.eigvalsh(H))
                K = np.linalg.cholesky(H - (min_eig - 1e-6) * np.eye(2 * n))
                warn(
                    f"WARNING: The system may be far from the magnetic ground-state. Minimum eigenvalue: {min_eig}. The magnon energies might be unphysical."
                )

        g = np.block([[1 * I, 0 * I], [0 * I, -1 * I]])
        KH = K.swapaxes(-1, -2).conj()

        return np.linalg.eigvalsh(KH @ g @ K)[:, n:] + min_eig

    def get_magnon_bands(
        self,
        kpoints: np.array = np.array([]),
        path: str = None,
        npoints: int = 300,
        special_points: dict = None,
        tol: float = 2e-4,
        pbc: tuple = None,
        cartesian: bool = False,
        labels: list = None,
        anisotropic: bool = True,
        u: np.array = None,
    ):
        """Get magnon band structure"""
        pbc = self._pbc if pbc is None else pbc
        u = self._uz if u is None else u
        if kpoints.size == 0:
            from ase.cell import Cell

            bandpath = Cell(self._cell).bandpath(
                path=path,
                npoints=npoints,
                special_points=special_points,
                eps=tol,
                pbc=pbc,
            )
            kpoints = bandpath.kpts
            spk = bandpath.special_points
            spk[r"$\Gamma$"] = spk.pop("G", np.zeros(3))
            labels = [
                (i, symbol)
                for symbol in spk
                for i in np.where((kpoints == spk[symbol]).all(axis=1))[0]
            ]
        elif cartesian:
            kpoints = np.linalg.solve(self._cell.T, kpoints.T).T

        bands = self._magnon_energies(kpoints, anisotropic=anisotropic)

        return labels, bands

    def plot_magnon_bands(self, **kwargs):
        """
        Plot magnon band structure.

        Parameters
        ----------
        **kwargs
            Additional keyword arguments passed to get_magnon_bands and plotting functions
        """
        filename = kwargs.pop("filename", None)
        kpath, bands = self.get_magnon_bands(**kwargs)
        bands_plot = BandsPlot(bands, kpath)
        bands_plot.plot(filename=filename)

    @classmethod
    def load_from_io(cls, exc: SpinIO, **kwargs):
        """
        Create Magnon instance from SpinIO

        Parameters
        ----------
        exc : SpinIO
            SpinIO instance with exchange parameters
        **kwargs : dict
            Additional arguments passed to get_full_Jtensor_for_Rlist

        Returns
        -------
        Magnon
            Initialized Magnon instance
        """
        return cls(
            nspin=exc.nspin,
            magmom=exc.magmoms,
            Rlist=exc.Rlist,
            JR=exc.get_full_Jtensor_for_Rlist(order="ij33", **kwargs),
        )

    @classmethod
    def from_TB2J_results(cls, path=None, fname="TB2J.pickle", **kwargs):
        """
        Create Magnon instance from TB2J results.

        Parameters
        ----------
        path : str, optional
            Path to the TB2J results file
        fname : str, optional
            Filename of the TB2J results file, default "TB2J.pickle"
        **kwargs : dict
            Additional arguments passed to load_from_io

        Returns
        -------
        Magnon
            Initialized Magnon instance
        """
        exc = SpinIO.load_pickle(path=path, fname=fname)
        return cls.load_from_io(exc, **kwargs)


def test_magnon(path="TB2J_results"):
    """Test the magnon calculator by loading from TB2J_results and computing at high-symmetry points."""
    from pathlib import Path

    import numpy as np

    # Check if TB2J_results exists
    results_path = Path(path)
    if not results_path.exists():
        raise FileNotFoundError(f"TB2J_results directory not found at {path}")

    # Load magnon calculator from TB2J results
    print(f"Loading exchange parameters from {path}...")
    magnon = Magnon.from_TB2J_results(path=path, iso_only=True)

    # Define high-symmetry points for a cube
    kpoints = np.array(
        [
            [0.0, 0.0, 0.0],  # Γ (Gamma)
            [0.5, 0.0, 0.0],  # X
            [0.5, 0.5, 0.0],  # M
            [0.5, 0.5, 0.5],  # R
        ]
    )
    klabels = ["Gamma", "X", "M", "R"]

    print("\nComputing exchange interactions at high-symmetry points...")
    Jq = magnon.Jq(kpoints)

    print(f"\nResults for {len(kpoints)} k-points:")
    print("-" * 50)
    print("Exchange interactions J(q):")
    print(f"Shape of Jq tensor: {Jq.shape}")
    print(
        f"Dimensions: (n_kpoints={Jq.shape[0]}, n_spin={Jq.shape[1]}, n_spin={Jq.shape[2]}, xyz={Jq.shape[3]}, xyz={Jq.shape[4]})"
    )

    print("\nComputing magnon energies...")
    energies = magnon._magnon_energies(kpoints)

    print("\nMagnon energies at high-symmetry points (in meV):")
    print("-" * 50)
    for i, (k, label) in enumerate(zip(kpoints, klabels)):
        print(f"\n{label}-point k={k}:")
        print(f"Energies: {energies[i] * 1000:.3f} meV")  # Convert to meV

    return magnon, Jq, energies


if __name__ == "__main__":
    test_magnon()



from .abacus import gen_exchange_abacus
from .manager import Manager
from .siesta_interface import gen_exchange_siesta
from .wannier90_interface import WannierManager, gen_exchange

__all__ = [
    "Manager",
    "gen_exchange_siesta",
    "WannierManager",
    "gen_exchange",
    "gen_exchange_abacus",
]



#!/usr/bin/env python3
import argparse
from TB2J.rotate_siestaDM import rotate_DM

def main():
    parser = argparse.ArgumentParser(description="")
    parser.add_argument(
        "--fdf_fname", help="Name of the *.fdf siesta file."
    )
    parser.add_argument(
        "--noncollinear",
        action="store_true",
        help="If present, six different configurations will be generated. These are required for non-collinear systems."
    )

    args = parser.parse_args()
    rotate_DM(args.fdf_fname, noncollinear=args.noncollinear)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import argparse
import sys

from TB2J.exchange_params import add_exchange_args_to_parser
from TB2J.interfaces import gen_exchange_siesta
from TB2J.versioninfo import print_license


def run_siesta2J():
    print_license()
    parser = argparse.ArgumentParser(
        description="siesta2J: Using magnetic force theorem to calculate exchange parameter J from siesta Hamiltonian"
    )
    # Add siesta specific arguments
    parser.add_argument(
        "--fdf_fname", help="path of the input fdf file", default="./", type=str
    )
    parser.add_argument(
        "--fname",
        default="exchange.xml",
        type=str,
        help="exchange xml file name. default: exchange.xml",
    )
    parser.add_argument(
        "--split_soc",
        help="whether the SOC part of the Hamiltonian can be read from the output of siesta. Default: False",
        action="store_true",
        default=False,
    )

    # Add common exchange arguments
    parser = add_exchange_args_to_parser(parser)

    args = parser.parse_args()

    if args.elements is None:
        print("Please input the magnetic elements, e.g. --elements Fe Ni")
        sys.exit()

    # include_orbs = {}
    # for element in args.elements:
    #    if "_" in element:
    #        elem = element.split("_")[0]
    #        orb = element.split("_")[1:]
    #        include_orbs[elem] = orb
    #    else:
    #        include_orbs[element] = None

    index_magnetic_atoms = args.index_magnetic_atoms
    if index_magnetic_atoms is not None:
        index_magnetic_atoms = [i - 1 for i in index_magnetic_atoms]

    gen_exchange_siesta(
        fdf_fname=args.fdf_fname,
        kmesh=args.kmesh,
        # magnetic_elements=list(include_orbs.keys()),
        # include_orbs=include_orbs,
        magnetic_elements=args.elements,
        include_orbs={},
        Rcut=args.rcut,
        emin=args.emin,
        emax=args.emax,
        nz=args.nz,
        description=args.description,
        output_path=args.output_path,
        use_cache=args.use_cache,
        nproc=args.np,
        exclude_orbs=args.exclude_orbs,
        orb_decomposition=args.orb_decomposition,
        read_H_soc=args.split_soc,
        orth=args.orth,
        index_magnetic_atoms=index_magnetic_atoms,
    )


if __name__ == "__main__":
    run_siesta2J()



# -*- coding: utf-8 -*-
"""
scan the spheres in the space of (theta, phi), so that the points are uniformly distributed on a sphere.
The algorithm is based on fibonacci spiral sampling method.
Reference:
https://extremelearning.com.au/how-to-evenly-distribute-points-on-a-sphere-more-effectively-than-the-canonical-fibonacci-lattice/

But note that the convention of theta and phi are different from the one in the reference.
Here, cos(theta) = z, and phi is the azimuthal angle in the x-y plane.
In the reference, theta is the azimuthal angle in the x-y plane, and phi is the polar angle.
"""

import numpy as np
from numpy import pi


def fibonacci_sphere(samples=100):
    """
    Fibonacci Sphere Sampling Method
    Parameters:
        samples (int): number of points to sample on the sphere
    Returns:
        theta and phi: numpy arrays with shape (samples,) containing the angles of the sampled points.
    """
    # Initialize the golden ratio and angles
    goldenRatio = (1 + np.sqrt(5)) / 2
    phi = np.arange(samples) * (2 * pi / goldenRatio)
    theta = np.arccos(1 - 2 * np.arange(samples) / samples)
    return theta, phi


def fibonacci_semisphere(samples=100):
    """
    Fibonacci Sphere Sampling Method for the upper hemisphere of a sphere.

    Parameters:
        samples (int): number of points to sample on the sphere

    Returns:
        theta and phi: numpy arrays with shape (samples,) containing the angles of the sampled points.
    """
    # Initialize the golden ratio and angles
    goldenRatio = (1 + np.sqrt(5)) / 2
    phi = np.arange(samples) * (2 * pi / goldenRatio)
    theta = np.arccos(np.linspace(0, 1, samples))
    return theta, phi


def test_fibonacci_sphere():
    import matplotlib.pyplot as plt

    # Generate points on the sphere
    samples = 200
    theta, phi = fibonacci_sphere(samples)
    # theta, phi = fibonacci_semisphere(samples)

    # Convert spherical coordinates to Cartesian coordinates
    x = np.cos(phi) * np.sin(theta)
    y = np.sin(phi) * np.sin(theta)
    z = np.cos(theta)

    fig = plt.figure()
    ax = fig.add_subplot(111, projection="3d")
    ax.scatter(x, y, z, s=13.5)
    # ax.plot(x, y, z)

    # Set aspect to 'equal' for equal scaling in all directions
    ax.set_aspect("equal")

    plt.show()


if __name__ == "__main__":
    test_fibonacci_sphere()



import numpy as np
import scipy as sp
from scipy.spatial.transform import Rotation
from numpy.linalg import norm, det


# Rotation from a to b
Rxx = Rotation.from_euler("x", 0, degrees=True)
Rxy = Rotation.from_euler("z", 90, degrees=True)
Rxz = Rotation.from_euler("y", -90, degrees=True)
# rotation from y to z
Ryx = Rotation.from_euler("z", -90, degrees=True)
Ryy = Rotation.from_euler("x", 0, degrees=True)
Ryz = Rotation.from_euler("x", 90, degrees=True)

# Inverse of the Rotation from x to z
Rzx = Rotation.from_euler("y", 90, degrees=True)
# Inverse of the Rotation from y to z
Rzy = Rotation.from_euler("x", -90, degrees=True)
Rzz = Rotation.from_euler("z", 0, degrees=True)


def normalize(vec):
    """
    Normalize vector
    :param vec: 3D vector
    :return: normalized vector
    """
    return vec / norm(vec)


def find_rotation_matrix_to_vec2(vec1, vec2=np.array([0, 0, 1])):
    """
    Find rotation matrix to align vector with vec2
    :param vec: 3D vector
    :param vec2: 3D vector
    :return: rotation matrix
    """
    # Normalize vector
    v1 = vec1 / norm(vec1)
    v2 = vec2 / norm(vec2)
    # Find rotation axis
    axis = np.cross(v1, np.array(v2))
    if norm(axis) < 1e-8:
        return np.eye(3, dtype=float)
    axis = axis / norm(axis)
    # Find rotation angle
    angle = np.arccos(np.dot(v1, v2))
    # Find rotation matrix
    rot = Rotation.from_rotvec(angle * axis)
    return rot.as_matrix()


def rotate_vec_to_vec2(vec, vec2=np.array([0, 0, 1])):
    """
    Rotate vector to align with z-axis
    :param vec: 3D vector
    :return: rotated vector
    """
    rot = find_rotation_matrix_to_vec2(vec, vec2)
    return np.dot(rot, vec)


def rotate_tensor(vec1, J, vec2):
    """
    Let H= vec1@T@vec2, rotate vec1 and vec2 to the z-axis, and return the rotated tensor
    """
    z = np.array([0, 0, 1])
    rot1 = find_rotation_matrix_to_vec2(vec1, z)
    rot2 = find_rotation_matrix_to_vec2(vec2, z)
    return rot1, rot2, rot1 @ J @ rot2.T


def rotate_back(rot1, rot2, J_rotated):
    """
    Rotate back the rotated tensor
    """
    return rot1.T @ J_rotated @ rot2


def remove_components(J, vec1, vec2, remove_indices=[[2, 2]]):
    """
    Remove some  component of the roated tensor and rotate back.
    :param J: tensor
    :param vec1: vector 1 which the tensor acts on (vec1@J@vec2)
    :param vec2: vector 2 which the tensor acts on (vec1@J@vec2)
    :param remove_indices: list of indices to remove, e.g. [[2,2]] to remove the zz component.
    Returns:
        Jback: the tensor with  removed components
        weigth_back: the weight of the Jback.
    """
    z = np.array([0, 0, 1])
    w = np.ones_like(J)
    rot1, rot2, J_rotated = rotate_tensor(vec1, J, vec2)
    w_rotated = rot1 @ w @ rot2.T
    for i, j in remove_indices:
        w_rotated[i, j] = 0
        J_rotated[i, j] = 0
    Jback = rotate_back(rot1, rot2, J_rotated)
    weight_back = rotate_back(rot1, rot2, w_rotated)
    # if np.linalg.norm(J)>1e-4:
    #    print(f"{vec1=}, {vec2=}")
    #    print(f"{J=}")
    #    print(f"{J_rotated=}")
    #    print(f"{Jback=}")
    #    print(f"{w_rotated=}")
    #    print(f"{weight_back=}")
    return Jback, weight_back


def get_weight_back(vec1, vec2):
    z = np.array([0, 0, 1])
    rot1 = find_rotation_matrix_to_vec2(vec1, z)
    rot2 = find_rotation_matrix_to_vec2(vec2, z)
    w = np.ones((3, 3))
    # w=np.eye(3)
    w_rotated = rot1 @ w @ rot2.T
    w_rotated[:, 2] = 0
    w_rotated[2, :] = 0
    w_rotated[2, 2] = 0
    return rot1.T @ w_rotated @ rot2


def test_weight_back():
    vec1 = [0.0, 0.0, 0.1]
    vec2 = [0.0, 0.0, 0.1]
    vec1 = normalize(vec1)
    vec2 = normalize(vec2)
    w = 0
    # Rotations=[Rxx, Rxy, Rxz, Ryx, Ryy, Ryz, Rzx, Rzy, Rzz]
    # wR=[0, 1, 1, 0, 0, 1, 0, 0, 0]
    # Rotations=[ Rxy, Rxz, Ryx,  Ryz, Rzx, Rzy]
    Rotations = [Rxy, Ryz, Rzx]
    # Rotations=[ Ryx, Rzy, Rxz]
    wR = [1] * len(Rotations)
    # Rotations=[Rxy, Ryz, Rzx]
    for wR, R in zip(wR, Rotations):
        v1 = R.apply(vec1)
        v2 = R.apply(vec2)
        wi = get_weight_back(v1, v2)
        print(wi * wR)
        w += wi * wR
    print((w + w.T) / 2)


# test_weight_back()


def test_remove_component():
    # vec1=[0.3, 0.4, 0.5]
    # vec2=[0.3, 0.4, -0.5]
    vec1 = [0.0, 0.1, 0]
    vec2 = [0.0, 0.1, 0]
    z = np.array([0, 0, 1])
    J = [[0.0, 0.1, 0.2], [0.1, 0.0, 0.3], [0.2, 0.3, 0.0]]
    J = [[0.0, 0.1, 0.2], [-0.1, 0.0, 0.3], [-0.2, -0.3, 0.0]]
    print(f"{J=}")
    w = np.ones_like(J)
    # w = np.eye(3)
    rot1, rot2, J_rotated = rotate_tensor(vec1, J, vec2)
    J_rotated = rot1 @ J @ rot2.T
    w_rotated = rot1 @ w @ rot2.T
    m = np.zeros((3, 3))
    m[0, 0] = 1
    m[0, 1] = 1
    # w_rotated = np.ones_like(J)
    print(f"{w_rotated=}")
    print(f"{J_rotated=}")
    print(f"{J*w_rotated=}")
    m = np.linalg.inv(J_rotated)
    remove = False
    if remove:
        w_rotated[:, 2] = 0
        w_rotated[2, :] = 0
        J_rotated[2, :] = 0
        J_rotated[:, 2] = 0
    m = m @ J_rotated
    Jback = rotate_back(rot1, rot2, J_rotated)
    # weight_back = rotate_back(rot1, rot2, w_rotated)
    weight_back = rot2 @ m @ rot2.T
    print(f"{weight_back=}")
    print(f"{Jback=}")
    print(f"{J@weight_back=}")
    print(f"{np.sum(weight_back)=}")


# test_remove_component()


def remove_zz_component(J, vec1, vec2):
    """
    Remove the zz component of the tensor
    """
    # return remove_components(J, vec1, vec2, remove_indices=[[2, 2]])
    return remove_components(
        J, vec1, vec2, remove_indices=[[0, 2], [1, 2], [2, 0], [2, 1], [2, 2]]
    )


def test_remove_zz_component():
    """
    Test remove_zz_component
    """
    J = np.array([[1, 2, 3], [2, 4, 5], [3, 5, 6]])
    vec1 = np.array([1, 0, 0])
    vec2 = np.array([1, 0, 0])
    J2, weight = remove_zz_component(J, vec1, vec2)
    print(J2)
    print(weight)
    print(vec1 @ J @ vec2)
    print(vec1 @ J2 @ vec2)
    print(vec1 @ weight @ vec2)
    # print(vec1@J2@vec2/vec1@weight@vec2)

    # Test with random vectors
    for i in range(10):
        vec1 = np.random.rand(3)
        vec2 = np.random.rand(3)
        J = np.random.rand(3, 3)
        J2, weight = remove_zz_component(J, vec1, vec2)
        # assert np.allclose(vec1@J@vec2, vec1@J2@vec2)
        # assert np.allclose(vec1@weight@vec2, 0)
        # assert np.allclose(weight[2, 2], 0)


def test_rotate_tensor():
    """
    Test rotate_tensor
    """
    J = np.array([[1, 2, 3], [2, 4, 5], [3, 5, 6]])
    vec1 = np.array([1, 0, 0])
    vec2 = np.array([0, 1, 0])
    rot1, rot2, J2 = rotate_tensor(vec1, J, vec2)
    print(J2)
    print(vec1 @ J @ vec2)
    print(vec1 @ rot1.T @ J2 @ rot2 @ vec2)

    # Test with random vectors
    for i in range(10):
        vec1 = np.random.rand(3)
        vec2 = np.random.rand(3)
        rot1, rot2, J2 = rotate_tensor(vec1, J, vec2)
        norm1 = norm(vec1)
        norm2 = norm(vec2)
        assert np.allclose(np.linalg.norm(vec1 @ rot1.T), norm(vec1))
        assert np.allclose(vec1 @ J @ vec2, vec1 @ rot1.T @ J2 @ rot2 @ vec2)
        assert np.allclose(rot1 @ rot1.T, np.eye(3))
        assert np.allclose(rot2 @ rot2.T, np.eye(3))


def test_find_rotation_matrix_to_z():
    """
    Test find_rotation_matrix_to_z
    """
    v = rotate_vec_to_vec2(np.array([1, 0, 0]), np.array([0, 0, 1]))
    assert np.allclose(v, np.array([0, 0, 1]))

    v = rotate_vec_to_vec2(np.array([-1, 0, 0]), np.array([0, 0, 1]))
    assert np.allclose(v, np.array([0, 0, 1]))

    v = rotate_vec_to_vec2(np.array([-1.0, 3, 8]), np.array([1, 0, 0]))
    print(v)
    assert np.allclose(v, np.array([1, 0, 0]))

    v = rotate_vec_to_vec2(np.array([-1.0, 3, 8]), np.array([3, 0.04, -0.3]))
    print(v)
    assert np.allclose(v, np.array([3, 0.04, -0.3]))


def test_rotate_isotropic_tensor():
    a = np.eye(3)
    b = rotate_tensor([0, 0, 1], np.eye(3), [1, 0, 0])
    print(b)


if __name__ == "__main__":
    # test_rotate_tensor()
    # test_find_rotation_matrix_to_z()
    # test_remove_zz_component()
    # test_rotate_isotropic_tensor()
    pass
